import dagster.check as check

from dagster import (
    DependencyDefinition,
    InputDefinition,
    Int,
    OutputDefinition,
    PipelineDefinition,
    Result,
    SolidDefinition,
    SolidInstance,
    execute_pipeline,
    execute_pipeline_iterator,
    lambda_solid,
    solid,
)

from dagster.core.types import Nullable, List, String

from dagster.core.definitions import Solid, solids_in_topological_order

from dagster.core.definitions.dependency import DependencyStructure

from dagster.core.definitions.pipeline import _create_adjacency_lists

from dagster.core.execution import PipelineExecutionResult, SolidExecutionResult

from dagster.core.utility_solids import define_stub_solid

from dagster.utils.test import execute_solid

# protected members
# pylint: disable=W0212


def _default_passthrough_transform(*args, **kwargs):
    check.invariant(not args, 'There should be no positional args')
    return list(kwargs.values())[0]


def create_dep_input_fn(name):
    return lambda context, arg_dict: {name: 'input_set'}


def make_transform():
    def transform(context, inputs):
        passed_rows = []
        seen = set()
        for row in inputs.values():
            for item in row:
                key = list(item.keys())[0]
                if key not in seen:
                    seen.add(key)
                    passed_rows.append(item)

        result = []
        result.extend(passed_rows)
        result.append({context.solid.name: 'transform_called'})
        return result

    return transform


def _transform_fn(context, inputs):
    passed_rows = []
    seen = set()
    for row in inputs.values():
        for item in row:
            key = list(item.keys())[0]
            if key not in seen:
                seen.add(key)
                passed_rows.append(item)

    result = []
    result.extend(passed_rows)
    result.append({context.solid.name: 'transform_called'})
    yield Result(result)


def create_solid_with_deps(name, *solid_deps):
    inputs = [InputDefinition(solid_dep.name) for solid_dep in solid_deps]

    return SolidDefinition(
        name=name, inputs=inputs, transform_fn=_transform_fn, outputs=[OutputDefinition()]
    )


def create_root_solid(name):
    input_name = name + '_input'
    inp = InputDefinition(input_name)

    return SolidDefinition(
        name=name, inputs=[inp], transform_fn=_transform_fn, outputs=[OutputDefinition()]
    )


def _do_construct(solids, dependencies):
    solids = {s.name: Solid(name=s.name, definition=s) for s in solids}
    dependency_structure = DependencyStructure.from_definitions(solids, dependencies)
    return _create_adjacency_lists(list(solids.values()), dependency_structure)


def test_empty_adjaceny_lists():
    solids = [create_root_solid('a_node')]
    forward_edges, backwards_edges = _do_construct(solids, {})
    assert forward_edges == {'a_node': set()}
    assert backwards_edges == {'a_node': set()}


def test_single_dep_adjacency_lists():
    # A <-- B
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    forward_edges, backwards_edges = _do_construct(
        [node_a, node_b], {'B': {'A': DependencyDefinition('A')}}
    )

    assert forward_edges == {'A': {'B'}, 'B': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set()}


def test_diamond_deps_adjaceny_lists():
    forward_edges, backwards_edges = _do_construct(create_diamond_solids(), diamond_deps())

    assert forward_edges == {'A_source': {'A'}, 'A': {'B', 'C'}, 'B': {'D'}, 'C': {'D'}, 'D': set()}
    assert backwards_edges == {
        'D': {'B', 'C'},
        'B': {'A'},
        'C': {'A'},
        'A': {'A_source'},
        'A_source': set(),
    }


def diamond_deps():
    return {
        'A': {'A_input': DependencyDefinition('A_source')},
        'B': {'A': DependencyDefinition('A')},
        'C': {'A': DependencyDefinition('A')},
        'D': {'B': DependencyDefinition('B'), 'C': DependencyDefinition('C')},
    }


def test_disconnected_graphs_adjaceny_lists():
    # A <-- B
    # C <-- D
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)

    node_c = create_root_solid('C')
    node_d = create_solid_with_deps('D', node_c)

    forward_edges, backwards_edges = _do_construct(
        [node_a, node_b, node_c, node_d],
        {'B': {'A': DependencyDefinition('A')}, 'D': {'C': DependencyDefinition('C')}},
    )
    assert forward_edges == {'A': {'B'}, 'B': set(), 'C': {'D'}, 'D': set()}
    assert backwards_edges == {'B': {'A'}, 'A': set(), 'D': {'C'}, 'C': set()}


def create_diamond_solids():
    a_source = define_stub_solid('A_source', [input_set('A_input')])
    node_a = create_root_solid('A')
    node_b = create_solid_with_deps('B', node_a)
    node_c = create_solid_with_deps('C', node_a)
    node_d = create_solid_with_deps('D', node_b, node_c)
    return [node_d, node_c, node_b, node_a, a_source]


def create_diamond_pipeline():
    return PipelineDefinition(
        name='diamond_pipeline', solids=create_diamond_solids(), dependencies=diamond_deps()
    )


def test_diamond_toposort():
    assert [s.name for s in solids_in_topological_order(create_diamond_pipeline())] == [
        'A_source',
        'A',
        'B',
        'C',
        'D',
    ]


def input_set(name):
    return {name: 'input_set'}


def transform_called(name):
    return {name: 'transform_called'}


def assert_equivalent_results(left, right):
    check.inst_param(left, 'left', SolidExecutionResult)
    check.inst_param(right, 'right', SolidExecutionResult)

    assert left.success == right.success
    assert left.name == right.name
    assert left.solid.name == right.solid.name
    assert left.transformed_value == right.transformed_value


def assert_all_results_equivalent(expected_results, result_results):
    check.list_param(expected_results, 'expected_results', of_type=SolidExecutionResult)
    check.list_param(result_results, 'result_results', of_type=SolidExecutionResult)
    assert len(expected_results) == len(result_results)
    for expected, result in zip(expected_results, result_results):
        assert_equivalent_results(expected, result)


def test_pipeline_execution_graph_diamond():
    pipeline = PipelineDefinition(solids=create_diamond_solids(), dependencies=diamond_deps())
    return _do_test(pipeline, lambda: execute_pipeline_iterator(pipeline))


def test_execute_solid_in_diamond():
    solid_result = execute_solid(
        create_diamond_pipeline(), 'A', inputs={'A_input': [{'a key': 'a value'}]}
    )

    assert solid_result.success
    assert solid_result.transformed_value() == [{'a key': 'a value'}, {'A': 'transform_called'}]


def test_execute_aliased_solid_in_diamond():
    a_source = define_stub_solid('A_source', [input_set('A_input')])
    pipeline_def = PipelineDefinition(
        name='aliased_pipeline',
        solids=[a_source, create_root_solid('A')],
        dependencies={
            SolidInstance('A', alias='aliased'): {'A_input': DependencyDefinition(a_source.name)}
        },
    )

    solid_result = execute_solid(
        pipeline_def, 'aliased', inputs={'A_input': [{'a key': 'a value'}]}
    )

    assert solid_result.success
    assert solid_result.transformed_value() == [
        {'a key': 'a value'},
        {'aliased': 'transform_called'},
    ]


def test_create_pipeline_with_empty_solids_list():
    single_solid_pipeline = PipelineDefinition(solids=[], dependencies={})

    result = execute_pipeline(single_solid_pipeline)
    assert result.success


def test_singleton_pipeline():
    stub_solid = define_stub_solid('stub', [{'a key': 'a value'}])
    single_solid_pipeline = PipelineDefinition(solids=[stub_solid], dependencies={})

    result = execute_pipeline(single_solid_pipeline)
    assert result.success


def test_two_root_solid_pipeline_with_empty_dependency_definition():
    stub_solid_a = define_stub_solid('stub_a', [{'a key': 'a value'}])
    stub_solid_b = define_stub_solid('stub_b', [{'a key': 'a value'}])
    single_solid_pipeline = PipelineDefinition(solids=[stub_solid_a, stub_solid_b], dependencies={})

    result = execute_pipeline(single_solid_pipeline)
    assert result.success


def test_two_root_solid_pipeline_with_partial_dependency_definition():
    stub_solid_a = define_stub_solid('stub_a', [{'a key': 'a value'}])
    stub_solid_b = define_stub_solid('stub_b', [{'a key': 'a value'}])
    single_solid_pipeline = PipelineDefinition(
        solids=[stub_solid_a, stub_solid_b], dependencies={'stub_a': {}}
    )

    result = execute_pipeline(single_solid_pipeline)
    assert result.success


def _do_test(pipeline, do_execute_pipeline_iter):

    results = list()

    for result in do_execute_pipeline_iter():
        results.append(result)

    result = PipelineExecutionResult(pipeline, 'kdjfkdjfd', results)

    assert result.result_for_solid('A').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
    ]

    assert result.result_for_solid('B').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('B'),
    ]

    assert result.result_for_solid('C').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('C'),
    ]

    assert result.result_for_solid('D').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('C'),
        transform_called('B'),
        transform_called('D'),
    ] or result.result_for_solid('D').transformed_value() == [
        input_set('A_input'),
        transform_called('A'),
        transform_called('B'),
        transform_called('C'),
        transform_called('D'),
    ]


def test_empty_pipeline_execution():
    result = execute_pipeline(PipelineDefinition(solids=[]))

    assert result.success


def test_pipeline_name_threaded_through_context():
    name = 'foobar'

    @solid()
    def assert_name_transform(context):
        assert context.pipeline_def.name == name

    result = execute_pipeline(PipelineDefinition(name="foobar", solids=[assert_name_transform]))

    assert result.success

    for step_event in execute_pipeline_iterator(
        PipelineDefinition(name="foobar", solids=[assert_name_transform]), {}
    ):
        assert step_event.is_step_success


def test_pipeline_subset():
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(inputs=[InputDefinition('num')])
    def add_one(num):
        return num + 1

    pipeline_def = PipelineDefinition(
        solids=[return_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('return_one')}},
    )

    pipeline_result = execute_pipeline(pipeline_def)
    assert pipeline_result.success
    assert pipeline_result.result_for_solid('add_one').transformed_value() == 2

    env_config = {'solids': {'add_one': {'inputs': {'num': {'value': 3}}}}}

    subset_result = execute_pipeline(
        pipeline_def.build_sub_pipeline(['add_one']), environment_dict=env_config
    )

    assert subset_result.success
    assert len(subset_result.solid_result_list) == 1
    assert subset_result.result_for_solid('add_one').transformed_value() == 4

    step_events = execute_pipeline_iterator(
        pipeline_def.build_sub_pipeline(['add_one']), environment_dict=env_config
    )

    for step_event in step_events:
        assert step_event.is_step_success


def define_three_part_pipeline():
    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_two(num):
        return num + 2

    @lambda_solid(inputs=[InputDefinition('num', Int)], output=OutputDefinition(Int))
    def add_three(num):
        return num + 3

    return PipelineDefinition(name='three_part_pipeline', solids=[add_one, add_two, add_three])


def define_created_disjoint_three_part_pipeline():
    return define_three_part_pipeline().build_sub_pipeline(['add_one', 'add_three'])


def test_pipeline_disjoint_subset():
    disjoint_pipeline = define_three_part_pipeline().build_sub_pipeline(['add_one', 'add_three'])
    assert len(disjoint_pipeline.solids) == 2


def test_pipeline_execution_disjoint_subset():
    env_config = {
        'solids': {
            'add_one': {'inputs': {'num': {'value': 2}}},
            'add_three': {'inputs': {'num': {'value': 5}}},
        },
        'context': {'default': {'config': {'log_level': 'ERROR'}}},
    }

    pipeline_def = define_created_disjoint_three_part_pipeline()

    result = execute_pipeline(
        pipeline_def.build_sub_pipeline(['add_one', 'add_three']), environment_dict=env_config
    )

    assert result.success

    assert len(result.solid_result_list) == 2

    assert result.result_for_solid('add_one').transformed_value() == 3
    assert result.result_for_solid('add_three').transformed_value() == 8


def test_pipeline_wrapping_types():
    @lambda_solid(
        inputs=[InputDefinition('value', Nullable(List(Nullable(String))))],
        output=OutputDefinition(Nullable(List(Nullable(String)))),
    )
    def double_string_for_all(value):
        if not value:
            return value

        output = []
        for item in value:
            output.append(None if item is None else item + item)
        return output

    pipeline_def = PipelineDefinition(name='wrapping_test', solids=[double_string_for_all])

    assert execute_pipeline(
        pipeline_def,
        environment_dict={'solids': {'double_string_for_all': {'inputs': {'value': None}}}},
    ).success

    assert execute_pipeline(
        pipeline_def,
        environment_dict={'solids': {'double_string_for_all': {'inputs': {'value': []}}}},
    ).success

    assert execute_pipeline(
        pipeline_def,
        environment_dict={
            'solids': {'double_string_for_all': {'inputs': {'value': [{'value': 'foo'}]}}}
        },
    ).success

    assert execute_pipeline(
        pipeline_def,
        environment_dict={
            'solids': {'double_string_for_all': {'inputs': {'value': [{'value': 'bar'}, None]}}}
        },
    ).success


def test_pipeline_streaming_iterator():
    events = []

    @lambda_solid
    def push_one():
        events.append(1)
        return 1

    @lambda_solid(inputs=[InputDefinition('num')])
    def add_one(num):
        events.append(num + 1)
        return num + 1

    pipeline_def = PipelineDefinition(
        name='test_streaming_iterator',
        solids=[push_one, add_one],
        dependencies={'add_one': {'num': DependencyDefinition('push_one')}},
    )

    step_event_iterator = execute_pipeline_iterator(pipeline_def)

    push_one_step_event = next(step_event_iterator)
    assert push_one_step_event.is_successful_output
    assert push_one_step_event.step_output_data.get_value() == 1
    assert events == [1]

    add_one_step_event = next(step_event_iterator)
    assert add_one_step_event.is_successful_output
    assert add_one_step_event.step_output_data.get_value() == 2
    assert events == [1, 2]


def test_pipeline_streaming_multiple_outputs():
    events = []

    @solid(outputs=[OutputDefinition(Int, 'one'), OutputDefinition(Int, 'two')])
    def push_one_two(_context):
        events.append(1)
        yield Result(1, 'one')
        events.append(2)
        yield Result(2, 'two')

    pipeline_def = PipelineDefinition(
        name='test_streaming_iterator_multiple_outputs', solids=[push_one_two]
    )

    step_event_iterator = execute_pipeline_iterator(pipeline_def)

    one_output_step_event = next(step_event_iterator)
    assert one_output_step_event.is_successful_output
    assert one_output_step_event.step_output_data.get_value() == 1
    assert one_output_step_event.step_output_data.output_name == 'one'
    assert events == [1]

    two_output_step_event = next(step_event_iterator)
    assert two_output_step_event.is_successful_output
    assert two_output_step_event.step_output_data.get_value() == 2
    assert two_output_step_event.step_output_data.output_name == 'two'
    assert events == [1, 2]



from collections import namedtuple
from dagster import check

from dagster.utils import camelcase, single_item

from dagster.core.system_config.objects import (
    ContextConfig,
    EnvironmentConfig,
    ExecutionConfig,
    ExpectationsConfig,
    SolidConfig,
)

from dagster.core.types import Bool, Field, List, NamedDict, NamedSelector, Dict
from dagster.core.types.config import ConfigType, ConfigTypeAttributes
from dagster.core.types.default_applier import apply_default_values
from dagster.core.types.field_utils import check_opt_field_param, FieldImpl

from .context import PipelineContextDefinition
from .resource import ResourceDefinition
from .dependency import Solid, SolidInputHandle, DependencyStructure
from .solid import SolidDefinition


def SystemNamedDict(name, fields, description=None):
    return NamedDict(name, fields, description, ConfigTypeAttributes(is_system_config=True))


def SystemNamedSelector(name, fields, description=None):
    return NamedSelector(name, fields, description, ConfigTypeAttributes(is_system_config=True))


def _is_selector_field_optional(config_type):
    check.inst_param(config_type, 'config_type', ConfigType)
    if len(config_type.fields) > 1:
        return False
    else:
        _name, field = single_item(config_type.fields)
        return field.is_optional


def define_maybe_optional_selector_field(config_cls):
    is_optional = _is_selector_field_optional(config_cls.inst())

    return (
        Field(
            config_cls,
            is_optional=is_optional,
            default_value=apply_default_values(config_cls.inst(), None),
        )
        if is_optional
        else Field(config_cls, is_optional=False)
    )


def define_resource_dictionary_cls(name, resources):
    check.str_param(name, 'name')
    check.dict_param(resources, 'resources', key_type=str, value_type=ResourceDefinition)

    fields = {}
    for resource_name, resource in resources.items():
        if resource.config_field:
            fields[resource_name] = Field(
                SystemNamedDict(name + '.' + resource_name, {'config': resource.config_field})
            )

    return SystemNamedDict(name=name, fields=fields)


def define_specific_context_config_cls(name, config_field, resources):
    check.str_param(name, 'name')
    check_opt_field_param(config_field, 'config_field')
    check.dict_param(resources, 'resources', key_type=str, value_type=ResourceDefinition)

    return SystemNamedDict(
        name,
        fields=remove_none_entries(
            {
                'config': config_field,
                'resources': Field(
                    define_resource_dictionary_cls('{name}.Resources'.format(name=name), resources)
                ),
                'persistence': Field(
                    SystemNamedSelector(
                        '{name}.Persistence'.format(name=name), {'file': Field(Dict({}))}
                    )
                ),
            }
        ),
    )


def define_context_context_cls(pipeline_name, context_definitions):
    check.str_param(pipeline_name, 'pipeline_name')
    check.dict_param(
        context_definitions,
        'context_definitions',
        key_type=str,
        value_type=PipelineContextDefinition,
    )
    full_type_name = '{pipeline_name}.ContextConfig'.format(pipeline_name=pipeline_name)
    field_dict = {}
    if len(context_definitions) == 1:
        context_name, context_definition = single_item(context_definitions)
        field_dict[context_name] = Field(
            define_specific_context_cls(pipeline_name, context_name, context_definition)
        )
    else:
        for context_name, context_definition in context_definitions.items():
            field_dict[context_name] = Field(
                define_specific_context_cls(pipeline_name, context_name, context_definition),
                is_optional=True,
            )

    return SystemNamedSelector(full_type_name, field_dict)


def define_specific_context_cls(pipeline_name, context_name, context_definition):
    return define_specific_context_config_cls(
        '{pipeline_name}.ContextDefinitionConfig.{context_name}'.format(
            pipeline_name=pipeline_name, context_name=camelcase(context_name)
        ),
        context_definition.config_field,
        context_definition.resources,
    )


def remove_none_entries(ddict):
    return {k: v for k, v in ddict.items() if v is not None}


def define_solid_config_cls(name, config_field, inputs_field, outputs_field):
    check.str_param(name, 'name')
    check_opt_field_param(config_field, 'config_field')
    check_opt_field_param(inputs_field, 'inputs_field')
    check_opt_field_param(outputs_field, 'outputs_field')

    return NamedDict(
        name,
        remove_none_entries(
            {'config': config_field, 'inputs': inputs_field, 'outputs': outputs_field}
        ),
        type_attributes=ConfigTypeAttributes(is_system_config=True),
    )


class EnvironmentClassCreationData(
    namedtuple(
        'EnvironmentClassCreationData',
        'pipeline_name solids context_definitions dependency_structure',
    )
):
    def __new__(cls, pipeline_name, solids, context_definitions, dependency_structure):
        return super(EnvironmentClassCreationData, cls).__new__(
            cls,
            check.str_param(pipeline_name, 'pipeline_name'),
            check.list_param(solids, 'solids', of_type=Solid),
            check.dict_param(
                context_definitions,
                'context_definitions',
                key_type=str,
                value_type=PipelineContextDefinition,
            ),
            check.inst_param(dependency_structure, 'dependency_structure', DependencyStructure),
        )


def define_context_cls(pipeline_def):
    from dagster import PipelineDefinition

    check.inst_param(pipeline_def, 'pipeline_def', PipelineDefinition)
    pipeline_name = camelcase(pipeline_def.name)
    return SystemNamedDict(
        name='{pipeline_name}.Context'.format(pipeline_name=pipeline_name),
        fields={
            'context': define_maybe_optional_selector_field(
                define_context_context_cls(pipeline_name, pipeline_def.context_definitions)
            )
        },
    )


def define_environment_cls(creation_data):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)
    pipeline_name = camelcase(creation_data.pipeline_name)
    return SystemNamedDict(
        name='{pipeline_name}.Environment'.format(pipeline_name=pipeline_name),
        fields={
            'context': define_maybe_optional_selector_field(
                define_context_context_cls(pipeline_name, creation_data.context_definitions)
            ),
            'solids': Field(
                define_solid_dictionary_cls(
                    '{pipeline_name}.SolidsConfigDictionary'.format(pipeline_name=pipeline_name),
                    creation_data,
                )
            ),
            'expectations': Field(
                define_expectations_config_cls(
                    '{pipeline_name}.ExpectationsConfig'.format(pipeline_name=pipeline_name)
                )
            ),
            'execution': Field(
                define_execution_config_cls(
                    '{pipeline_name}.ExecutionConfig'.format(pipeline_name=pipeline_name)
                )
            ),
        },
    )


def define_expectations_config_cls(name):
    check.str_param(name, 'name')

    return SystemNamedDict(
        name, fields={'evaluate': Field(Bool, is_optional=True, default_value=True)}
    )


def solid_has_configurable_inputs(solid_def):
    check.inst_param(solid_def, 'solid_def', SolidDefinition)
    return any(map(lambda inp: inp.runtime_type.input_schema, solid_def.input_defs))


def solid_has_configurable_outputs(solid_def):
    check.inst_param(solid_def, 'solid_def', SolidDefinition)
    return any(map(lambda out: out.runtime_type.output_schema, solid_def.output_defs))


def get_inputs_field(creation_data, solid):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)
    check.inst_param(solid, 'solid', Solid)

    if not solid_has_configurable_inputs(solid.definition):
        return None

    inputs_field_fields = {}
    for inp in [inp for inp in solid.definition.input_defs if inp.runtime_type.input_schema]:
        inp_handle = SolidInputHandle(solid, inp)
        # If this input is not satisfied by a dependency you must
        # provide it via config
        if not creation_data.dependency_structure.has_dep(inp_handle):
            inputs_field_fields[inp.name] = FieldImpl(inp.runtime_type.input_schema.schema_type)

    if not inputs_field_fields:
        return None

    return Field(
        SystemNamedDict(
            '{pipeline_name}.{solid_name}.Inputs'.format(
                pipeline_name=camelcase(creation_data.pipeline_name),
                solid_name=camelcase(solid.name),
            ),
            inputs_field_fields,
        )
    )


def get_outputs_field(creation_data, solid):
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)
    check.inst_param(solid, 'solid', Solid)

    solid_def = solid.definition

    if not solid_has_configurable_outputs(solid_def):
        return None

    output_dict_fields = {}
    for out in [out for out in solid_def.output_defs if out.runtime_type.output_schema]:
        output_dict_fields[out.name] = Field(
            type(out.runtime_type.output_schema.schema_type), is_optional=True
        )

    output_entry_dict = SystemNamedDict(
        '{pipeline_name}.{solid_name}.Outputs'.format(
            pipeline_name=camelcase(creation_data.pipeline_name), solid_name=camelcase(solid.name)
        ),
        output_dict_fields,
    )

    return Field(List(output_entry_dict), is_optional=True)


def solid_has_config_entry(solid_def):
    check.inst_param(solid_def, 'solid_def', SolidDefinition)
    return (
        solid_def.config_field
        or solid_has_configurable_inputs(solid_def)
        or solid_has_configurable_outputs(solid_def)
    )


def define_solid_dictionary_cls(name, creation_data):
    check.str_param(name, 'name')
    check.inst_param(creation_data, 'creation_data', EnvironmentClassCreationData)

    fields = {}
    for solid in creation_data.solids:
        if solid_has_config_entry(solid.definition):
            solid_config_type = define_solid_config_cls(
                '{pipeline_name}.SolidConfig.{solid_name}'.format(
                    pipeline_name=camelcase(creation_data.pipeline_name),
                    solid_name=camelcase(solid.name),
                ),
                solid.definition.config_field,
                inputs_field=get_inputs_field(creation_data, solid),
                outputs_field=get_outputs_field(creation_data, solid),
            )
            fields[solid.name] = Field(solid_config_type)

    return SystemNamedDict(name, fields)


def define_execution_config_cls(name):
    check.str_param(name, 'name')
    return NamedDict(name, {}, type_attributes=ConfigTypeAttributes(is_system_config=True))


def construct_environment_config(config_value):
    check.dict_param(config_value, 'config_value')
    return EnvironmentConfig(
        solids=construct_solid_dictionary(config_value['solids']),
        execution=ExecutionConfig(**config_value['execution']),
        expectations=ExpectationsConfig(**config_value['expectations']),
        context=construct_context_config(config_value['context']),
        original_config_dict=config_value,
    )


def construct_context_config(config_value):
    context_name, context_value = single_item(config_value)
    return ContextConfig(
        name=context_name,
        config=context_value.get('config'),
        resources=context_value['resources'],
        persistence=context_value['persistence'],
    )


def construct_solid_dictionary(solid_dict_value):
    return {
        key: SolidConfig(
            config=value.get('config'),
            inputs=value.get('inputs', {}),
            outputs=value.get('outputs', []),
        )
        for key, value in solid_dict_value.items()
    }



from collections import namedtuple
from enum import Enum
import imp
import importlib
import os

import click

from dagster import PipelineDefinition, RepositoryDefinition, check

from dagster.utils import load_yaml_from_path


INFO_FIELDS = set(['repository_yaml', 'pipeline_name', 'python_file', 'fn_name', 'module_name'])

PipelineTargetInfo = namedtuple(
    'PipelineTargetInfo', 'repository_yaml pipeline_name python_file fn_name module_name'
)


RepositoryTargetInfo = namedtuple(
    'RepositoryTargetInfo', 'repository_yaml module_name python_file fn_name'
)


class PipelineTargetMode(Enum):
    PIPELINE = 1
    REPOSITORY = 2


RepositoryData = namedtuple('RepositoryData', 'entrypoint pipeline_name')

PipelineLoadingModeData = namedtuple('PipelineLoadingModeData', 'mode data')


class InvalidPipelineLoadingComboError(Exception):
    pass


class InvalidRepositoryLoadingComboError(Exception):
    pass


def check_info_fields(pipeline_target_info, *fields):
    check.inst_param(pipeline_target_info, 'pipeline_target_info', PipelineTargetInfo)
    check.tuple_param(fields, 'fields')

    pipeline_target_dict = pipeline_target_info._asdict()
    for field in fields:
        if pipeline_target_dict[field] is None:
            return False

    for none_field in INFO_FIELDS.difference(set(fields)):
        if pipeline_target_dict[none_field] is not None:
            raise InvalidPipelineLoadingComboError(
                (
                    'field: {none_field} with value {value} should not be set if'
                    '{fields} were provided'
                ).format(
                    value=repr(pipeline_target_dict[none_field]),
                    none_field=none_field,
                    fields=repr(fields),
                )
            )

    return True


def repo_load_invariant(condition):
    if not condition:
        raise InvalidRepositoryLoadingComboError()


def entrypoint_from_repo_target_info(repo_target_info):
    check.inst_param(repo_target_info, 'repo_target_info', RepositoryTargetInfo)

    if repo_target_info.repository_yaml:
        repo_load_invariant(repo_target_info.module_name is None)
        repo_load_invariant(repo_target_info.python_file is None)
        repo_load_invariant(repo_target_info.fn_name is None)
        return entrypoint_from_yaml(repo_target_info.repository_yaml)
    elif repo_target_info.module_name and repo_target_info.fn_name:
        repo_load_invariant(repo_target_info.repository_yaml is None)
        repo_load_invariant(repo_target_info.python_file is None)
        return entrypoint_from_module_target(
            module_name=repo_target_info.module_name, fn_name=repo_target_info.fn_name
        )
    elif repo_target_info.python_file and repo_target_info.fn_name:
        repo_load_invariant(repo_target_info.repository_yaml is None)
        repo_load_invariant(repo_target_info.module_name is None)
        return entrypoint_from_file_target(
            python_file=repo_target_info.python_file, fn_name=repo_target_info.fn_name
        )
    else:
        raise InvalidRepositoryLoadingComboError()


def create_pipeline_loading_mode_data(pipeline_target_info):
    check.inst_param(pipeline_target_info, 'pipeline_target_info', PipelineTargetInfo)

    if check_info_fields(pipeline_target_info, 'python_file', 'fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY,
            data=RepositoryData(
                entrypoint=entrypoint_from_file_target(
                    python_file=pipeline_target_info.python_file,
                    fn_name=pipeline_target_info.fn_name,
                ),
                pipeline_name=pipeline_target_info.pipeline_name,
            ),
        )
    elif check_info_fields(pipeline_target_info, 'module_name', 'fn_name', 'pipeline_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY,
            data=RepositoryData(
                entrypoint=entrypoint_from_module_target(
                    module_name=pipeline_target_info.module_name,
                    fn_name=pipeline_target_info.fn_name,
                ),
                pipeline_name=pipeline_target_info.pipeline_name,
            ),
        )
    elif check_info_fields(pipeline_target_info, 'python_file', 'fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE,
            data=entrypoint_from_file_target(
                python_file=pipeline_target_info.python_file, fn_name=pipeline_target_info.fn_name
            ),
        )
    elif check_info_fields(pipeline_target_info, 'module_name', 'fn_name'):
        return PipelineLoadingModeData(
            mode=PipelineTargetMode.PIPELINE,
            data=entrypoint_from_module_target(
                module_name=pipeline_target_info.module_name, fn_name=pipeline_target_info.fn_name
            ),
        )
    elif pipeline_target_info.pipeline_name:
        for none_field in ['python_file', 'fn_name', 'module_name']:
            if getattr(pipeline_target_info, none_field) is not None:
                raise InvalidPipelineLoadingComboError(
                    '{none_field} is not None. Got {value}'.format(
                        none_field=none_field, value=getattr(pipeline_target_info, none_field)
                    )
                )

        check.invariant(pipeline_target_info.repository_yaml is not None)

        return PipelineLoadingModeData(
            mode=PipelineTargetMode.REPOSITORY,
            data=RepositoryData(
                entrypoint=entrypoint_from_yaml(pipeline_target_info.repository_yaml),
                pipeline_name=pipeline_target_info.pipeline_name,
            ),
        )
    else:
        raise InvalidPipelineLoadingComboError()


LoaderEntrypoint = namedtuple('LoaderEntrypoint', 'module module_name fn_name')


def perform_load(entry):
    fn = getattr(entry.module, entry.fn_name)
    check.is_callable(fn)
    return fn()


def entrypoint_from_file_target(python_file, fn_name):
    module_name = os.path.splitext(os.path.basename(python_file))[0]
    module = imp.load_source(module_name, python_file)
    return LoaderEntrypoint(module, module_name, fn_name)


def entrypoint_from_module_target(module_name, fn_name):
    module = importlib.import_module(module_name)
    return LoaderEntrypoint(module, module_name, fn_name)


EMPHERMAL_NAME = '<<unnamed>>'


def load_repository_from_target_info(repo_target_info):
    check.inst_param(repo_target_info, 'repo_target_info', RepositoryTargetInfo)
    entrypoint = entrypoint_from_repo_target_info(repo_target_info)
    obj = perform_load(entrypoint)

    if isinstance(obj, RepositoryDefinition):
        return obj
    elif isinstance(obj, PipelineDefinition):
        return RepositoryDefinition(name=EMPHERMAL_NAME, pipeline_dict={obj.name: lambda: obj})
    else:
        raise InvalidPipelineLoadingComboError('entry point must return a repository or pipeline')


def load_pipeline_from_target_info(pipeline_target_info):
    check.inst_param(pipeline_target_info, 'pipeline_target_info', PipelineTargetInfo)

    mode_data = create_pipeline_loading_mode_data(pipeline_target_info)

    if mode_data.mode == PipelineTargetMode.REPOSITORY:
        repository = check.inst(perform_load(mode_data.data.entrypoint), RepositoryDefinition)
        return repository.get_pipeline(mode_data.data.pipeline_name)
    elif mode_data.mode == PipelineTargetMode.PIPELINE:
        return check.inst(perform_load(mode_data.data), PipelineDefinition)
    else:
        check.failed('Should never reach')


def entrypoint_from_yaml(file_path):
    check.str_param(file_path, 'file_path')

    config = load_yaml_from_path(file_path)
    repository_config = check.dict_elem(config, 'repository')
    module_name = check.opt_str_elem(repository_config, 'module')
    file_name = check.opt_str_elem(repository_config, 'file')
    fn_name = check.str_elem(repository_config, 'fn')

    if module_name:
        return entrypoint_from_module_target(module_name, fn_name)
    else:
        # rebase file in config off of the path in the config file
        file_name = os.path.join(os.path.dirname(os.path.abspath(file_path)), file_name)
        return entrypoint_from_file_target(file_name, fn_name)


def apply_click_params(command, *click_params):
    for click_param in click_params:
        command = click_param(command)
    return command


def repository_target_argument(f):
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help=(
                'Path to config file. Defaults to ./repository.yml. if --python-file '
                'and --module-name are not specified'
            ),
        ),
        click.option(
            '--python-file',
            '-f',
            help='Specify python file where repository or pipeline function lives.',
        ),
        click.option(
            '--module-name', '-m', help='Specify module where repository or pipeline function lives'
        ),
        click.option('--fn-name', '-n', help='Function that returns either repository or pipeline'),
    )


def pipeline_target_command(f):
    # f = repository_config_argument(f)
    # nargs=-1 is used right now to make this argument optional
    # it can only handle 0 or 1 pipeline names
    # see .pipeline.create_pipeline_from_cli_args
    return apply_click_params(
        f,
        click.option(
            '--repository-yaml',
            '-y',
            type=click.STRING,
            help=(
                'Path to config file. Defaults to ./repository.yml. if --python-file '
                'and --module-name are not specified'
            ),
        ),
        click.argument('pipeline_name', nargs=-1),
        click.option('--python-file', '-f'),
        click.option('--module-name', '-m'),
        click.option('--fn-name', '-n'),
    )


def all_none(kwargs):
    for value in kwargs.values():
        if value is not None:
            return False
    return True


def load_target_info_from_cli_args(cli_args):
    check.dict_param(cli_args, 'cli_args')

    if all_none(cli_args):
        cli_args['repository_yaml'] = 'repository.yml'

    return RepositoryTargetInfo(**cli_args)



import pytest

from dagster import (
    Any,
    DagsterEvaluateConfigValueError,
    DependencyDefinition,
    Dict,
    ExecutionContext,
    Field,
    InputDefinition,
    Int,
    NamedDict,
    OutputDefinition,
    PipelineContextDefinition,
    PipelineDefinition,
    SolidDefinition,
    SolidInstance,
    String,
    execute_pipeline,
    lambda_solid,
    types,
)


from dagster.core.system_config.objects import (
    ContextConfig,
    EnvironmentConfig,
    ExpectationsConfig,
    SolidConfig,
)

from dagster.core.definitions.environment_configs import (
    construct_context_config,
    construct_environment_config,
    construct_solid_dictionary,
    define_context_context_cls,
    define_expectations_config_cls,
    define_solid_config_cls,
    define_solid_dictionary_cls,
    EnvironmentClassCreationData,
)

from dagster.core.types.evaluator import evaluate_config_value

from dagster.core.test_utils import throwing_evaluate_config_value


def create_creation_data(pipeline_def):
    return EnvironmentClassCreationData(
        pipeline_def.name,
        pipeline_def.solids,
        pipeline_def.context_definitions,
        pipeline_def.dependency_structure,
    )


def test_context_config_any():
    context_defs = {
        'test': PipelineContextDefinition(
            config_field=Field(Any), context_fn=lambda *args: ExecutionContext()
        )
    }

    context_config_type = define_context_context_cls('something', context_defs).inst()

    assert context_config_type.type_attributes.is_system_config

    output = construct_context_config(
        throwing_evaluate_config_value(context_config_type, {'test': {'config': 1}})
    )
    assert output.name == 'test'
    assert output.config == 1


def test_context_config():
    context_defs = {
        'test': PipelineContextDefinition(
            config_field=Field(dagster_type=Dict({'some_str': Field(String)})),
            context_fn=lambda *args: ExecutionContext(),
        )
    }

    context_config_type = define_context_context_cls('something', context_defs).inst()

    output = construct_context_config(
        throwing_evaluate_config_value(
            context_config_type, {'test': {'config': {'some_str': 'something'}}}
        )
    )
    assert isinstance(output, ContextConfig)
    assert output.name == 'test'
    assert output.config == {'some_str': 'something'}


def test_default_expectations():
    expect_config_type = define_expectations_config_cls('some_name').inst()
    assert (
        ExpectationsConfig(**throwing_evaluate_config_value(expect_config_type, {})).evaluate
        is True
    )
    assert (
        ExpectationsConfig(**throwing_evaluate_config_value(expect_config_type, None)).evaluate
        is True
    )


def test_default_context_config():
    pipeline_def = PipelineDefinition(
        solids=[
            SolidDefinition(
                name='some_solid', inputs=[], outputs=[], transform_fn=lambda *args: None
            )
        ]
    )

    context_config_type = define_context_context_cls(
        pipeline_def.name, pipeline_def.context_definitions
    ).inst()
    assert 'default' in context_config_type.fields
    assert context_config_type.fields['default'].is_optional
    default_context_config_type = context_config_type.fields['default'].config_type

    assert 'config' in default_context_config_type.fields

    context_dict = throwing_evaluate_config_value(context_config_type, {})

    assert 'default' in context_dict


def test_all_types_provided():
    pipeline_def = PipelineDefinition(
        name='pipeline',
        solids=[],
        context_definitions={
            'some_context': PipelineContextDefinition(
                config_field=Field(
                    NamedDict(
                        'SomeContextNamedDict',
                        {'with_default_int': Field(Int, is_optional=True, default_value=23434)},
                    )
                ),
                context_fn=lambda *args: None,
            )
        },
    )

    all_types = list(pipeline_def.all_config_types())
    type_names = set(t.name for t in all_types)
    assert 'SomeContextNamedDict' in type_names
    assert 'Pipeline.ContextDefinitionConfig.SomeContext' in type_names
    assert 'Pipeline.ContextDefinitionConfig.SomeContext.Resources' in type_names


def test_provided_default_config():
    pipeline_def = PipelineDefinition(
        context_definitions={
            'some_context': PipelineContextDefinition(
                config_field=Field(
                    Dict({'with_default_int': Field(Int, is_optional=True, default_value=23434)})
                ),
                context_fn=lambda *args: None,
            )
        },
        solids=[
            SolidDefinition(
                name='some_solid', inputs=[], outputs=[], transform_fn=lambda *args: None
            )
        ],
    )

    env_type = pipeline_def.environment_type
    some_context_field = env_type.fields['context'].config_type.fields['some_context']
    assert some_context_field.is_optional

    some_context_config_field = some_context_field.config_type.fields['config']
    assert some_context_config_field.is_optional
    assert some_context_config_field.default_value == {'with_default_int': 23434}

    assert some_context_field.default_value == {
        'config': {'with_default_int': 23434},
        'resources': {},
        'persistence': {'file': {}},
    }

    value = construct_environment_config(
        throwing_evaluate_config_value(pipeline_def.environment_type, {})
    )
    assert value.context.name == 'some_context'
    assert env_type.type_attributes.is_system_config


def test_default_environment():
    pipeline_def = PipelineDefinition(
        solids=[
            SolidDefinition(
                name='some_solid', inputs=[], outputs=[], transform_fn=lambda *args: None
            )
        ]
    )

    env_obj = construct_environment_config(
        throwing_evaluate_config_value(pipeline_def.environment_type, {})
    )

    assert env_obj.expectations.evaluate is True


def test_errors():
    context_defs = {
        'test': PipelineContextDefinition(
            config_field=Field(Dict({'required_int': Field(Int)})),
            context_fn=lambda *args: ExecutionContext(),
        )
    }

    context_config_type = define_context_context_cls('something', context_defs).inst()

    assert not evaluate_config_value(context_config_type, 1).success
    assert not evaluate_config_value(context_config_type, {}).success

    invalid_value = {'context_one': 1, 'context_two': 2}

    result = evaluate_config_value(context_config_type, invalid_value)
    assert not result.success
    assert len(result.errors) == 1


def test_select_context():
    context_defs = {
        'int_context': PipelineContextDefinition(
            config_field=Field(Int), context_fn=lambda *args: ExecutionContext()
        ),
        'string_context': PipelineContextDefinition(
            config_field=Field(String), context_fn=lambda *args: ExecutionContext()
        ),
    }

    context_config_type = define_context_context_cls('something', context_defs).inst()

    assert construct_context_config(
        throwing_evaluate_config_value(context_config_type, {'int_context': {'config': 1}})
    ) == ContextConfig(name='int_context', config=1)

    assert construct_context_config(
        throwing_evaluate_config_value(context_config_type, {'string_context': {'config': 'bar'}})
    ) == ContextConfig(name='string_context', config='bar')

    # mismatched field type mismatch
    with pytest.raises(DagsterEvaluateConfigValueError):
        assert throwing_evaluate_config_value(
            context_config_type, {'int_context': {'config': 'bar'}}
        )

    # mismatched field type mismatch
    with pytest.raises(DagsterEvaluateConfigValueError):
        assert throwing_evaluate_config_value(
            context_config_type, {'string_context': {'config': 1}}
        )


def test_solid_config():
    solid_config_type = define_solid_config_cls('kdjfkd', Field(Int), None, None).inst()
    solid_inst = throwing_evaluate_config_value(solid_config_type, {'config': 1})
    assert solid_inst['config'] == 1
    assert solid_config_type.inst().type_attributes.is_system_config


def test_expectations_config():
    expectations_config_type = define_expectations_config_cls('ksjdfkd').inst()
    expectations = ExpectationsConfig(
        **throwing_evaluate_config_value(expectations_config_type, {'evaluate': True})
    )

    assert isinstance(expectations, ExpectationsConfig)
    assert expectations.evaluate is True

    assert ExpectationsConfig(
        **throwing_evaluate_config_value(expectations_config_type, {'evaluate': False})
    ) == ExpectationsConfig(evaluate=False)


def test_solid_dictionary_type():
    pipeline_def = define_test_solids_config_pipeline()

    solid_dict_type = define_solid_dictionary_cls(
        'foobar', create_creation_data(pipeline_def)
    ).inst()

    value = construct_solid_dictionary(
        throwing_evaluate_config_value(
            solid_dict_type,
            {'int_config_solid': {'config': 1}, 'string_config_solid': {'config': 'bar'}},
        )
    )

    assert set(['int_config_solid', 'string_config_solid']) == set(value.keys())
    assert value == {'int_config_solid': SolidConfig(1), 'string_config_solid': SolidConfig('bar')}

    assert solid_dict_type.type_attributes.is_system_config

    for specific_solid_config_field in solid_dict_type.fields.values():
        specific_solid_config_type = specific_solid_config_field.config_type
        assert specific_solid_config_type.type_attributes.is_system_config
        user_config_field = specific_solid_config_field.config_type.fields['config']
        assert user_config_field.config_type.type_attributes.is_system_config is False


def define_test_solids_config_pipeline():
    return PipelineDefinition(
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int, is_optional=True),
                inputs=[],
                outputs=[],
                transform_fn=lambda *args: None,
            ),
            SolidDefinition(
                name='string_config_solid',
                config_field=Field(String, is_optional=True),
                inputs=[],
                outputs=[],
                transform_fn=lambda *args: None,
            ),
        ]
    )


def assert_has_fields(dtype, *fields):
    return set(dtype.fields.keys()) == set(fields)


def test_solid_configs_defaults():
    env_type = define_test_solids_config_pipeline().environment_type

    solids_field = env_type.fields['solids']

    assert_has_fields(solids_field.config_type, 'int_config_solid', 'string_config_solid')

    int_solid_field = solids_field.config_type.fields['int_config_solid']

    assert int_solid_field.is_optional
    # TODO: this is the test case the exposes the default dodginess
    assert int_solid_field.default_provided

    assert_has_fields(int_solid_field.config_type, 'config')

    int_solid_config_field = int_solid_field.config_type.fields['config']

    assert int_solid_config_field.is_optional
    assert not int_solid_config_field.default_provided


def test_solid_dictionary_some_no_config():
    pipeline_def = PipelineDefinition(
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                inputs=[],
                outputs=[],
                transform_fn=lambda *args: None,
            ),
            SolidDefinition(
                name='no_config_solid', inputs=[], outputs=[], transform_fn=lambda *args: None
            ),
        ]
    )

    solid_dict_type = define_solid_dictionary_cls(
        'foobar', create_creation_data(pipeline_def)
    ).inst()

    value = construct_solid_dictionary(
        throwing_evaluate_config_value(solid_dict_type, {'int_config_solid': {'config': 1}})
    )

    assert set(['int_config_solid']) == set(value.keys())
    assert value == {'int_config_solid': SolidConfig(1)}


def test_whole_environment():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        context_definitions={
            'test': PipelineContextDefinition(
                config_field=Field(Any), context_fn=lambda *args: ExecutionContext()
            )
        },
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                inputs=[],
                outputs=[],
                transform_fn=lambda *args: None,
            ),
            SolidDefinition(
                name='no_config_solid', inputs=[], outputs=[], transform_fn=lambda *args: None
            ),
        ],
    )

    environment_type = pipeline_def.environment_type

    assert environment_type.fields['context'].config_type.name == 'SomePipeline.ContextConfig'
    solids_type = environment_type.fields['solids'].config_type
    assert solids_type.name == 'SomePipeline.SolidsConfigDictionary'
    assert (
        solids_type.fields['int_config_solid'].config_type.name
        == 'SomePipeline.SolidConfig.IntConfigSolid'
    )
    assert (
        environment_type.fields['expectations'].config_type.name
        == 'SomePipeline.ExpectationsConfig'
    )

    env = construct_environment_config(
        throwing_evaluate_config_value(
            pipeline_def.environment_type,
            {'context': {'test': {'config': 1}}, 'solids': {'int_config_solid': {'config': 123}}},
        )
    )

    assert isinstance(env, EnvironmentConfig)
    assert env.context == ContextConfig('test', 1, persistence={'file': {}})
    assert env.solids == {'int_config_solid': SolidConfig(123)}
    assert env.expectations == ExpectationsConfig(evaluate=True)


def test_solid_config_error():
    solid_dict_type = define_solid_dictionary_cls(
        'slkdfjkjdsf', create_creation_data(define_test_solids_config_pipeline())
    ).inst()

    int_solid_config_type = solid_dict_type.fields['int_config_solid'].config_type

    with pytest.raises(DagsterEvaluateConfigValueError, match='Field "notconfig" is not defined'):
        throwing_evaluate_config_value(int_solid_config_type, {'notconfig': 1})

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(int_solid_config_type, 1)


def test_optional_solid_with_no_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                inputs=[],
                outputs=[],
                transform_fn=lambda context, _inputs: _assert_config_none(context, 234),
            ),
            SolidDefinition(
                name='no_config_solid',
                inputs=[],
                outputs=[],
                transform_fn=lambda context, _inputs: _assert_config_none(context, None),
            ),
        ],
    )

    assert execute_pipeline(pipeline_def, {'solids': {'int_config_solid': {'config': 234}}}).success


def test_optional_solid_with_optional_scalar_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int, is_optional=True),
                inputs=[],
                outputs=[],
                transform_fn=lambda context, _inputs: _assert_config_none(context, 234),
            )
        ],
    )

    env_type = pipeline_def.environment_type

    assert env_type.fields['solids'].is_optional is True

    solids_type = env_type.fields['solids'].config_type

    assert solids_type.fields['int_config_solid'].is_optional is True

    solids_default_obj = construct_solid_dictionary(throwing_evaluate_config_value(solids_type, {}))

    assert solids_default_obj['int_config_solid'].config is None

    env_obj = construct_environment_config(
        throwing_evaluate_config_value(pipeline_def.environment_type, {})
    )

    assert env_obj.solids['int_config_solid'].config is None


def test_optional_solid_with_required_scalar_config():
    def _assert_config_none(context, value):
        assert context.solid_config is value

    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Int),
                inputs=[],
                outputs=[],
                transform_fn=lambda context, _inputs: _assert_config_none(context, 234),
            )
        ],
    )

    env_type = pipeline_def.environment_type

    assert env_type.fields['solids'].is_optional is False

    solids_type = env_type.fields['solids'].config_type

    assert solids_type.fields['int_config_solid'].is_optional is False

    int_config_solid_type = solids_type.fields['int_config_solid'].config_type

    assert_has_fields(int_config_solid_type, 'config')

    int_config_solid_config_field = int_config_solid_type.fields['config']

    assert int_config_solid_config_field.is_optional is False

    execute_pipeline(pipeline_def, {'solids': {'int_config_solid': {'config': 234}}})


def test_required_solid_with_required_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(Dict({'required_field': Field(String)})),
                inputs=[],
                outputs=[],
                transform_fn=lambda *_args: None,
            )
        ],
    )

    env_type = pipeline_def.environment_type

    assert env_type.fields['solids'].is_optional is False
    assert env_type.fields['solids'].config_type

    solids_type = env_type.fields['solids'].config_type
    assert solids_type.fields['int_config_solid'].is_optional is False
    int_config_solid_type = solids_type.fields['int_config_solid'].config_type
    assert int_config_solid_type.fields['config'].is_optional is False

    assert env_type.fields['context'].is_optional
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['expectations'].is_optional

    env_obj = construct_environment_config(
        throwing_evaluate_config_value(
            pipeline_def.environment_type,
            {'solids': {'int_config_solid': {'config': {'required_field': 'foobar'}}}},
        )
    )

    assert env_obj.solids['int_config_solid'].config['required_field'] == 'foobar'

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(pipeline_def.environment_type, {'solids': {}})

    with pytest.raises(DagsterEvaluateConfigValueError):
        throwing_evaluate_config_value(pipeline_def.environment_type, {})


def test_optional_solid_with_optional_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[
            SolidDefinition(
                name='int_config_solid',
                config_field=Field(
                    Dict({'optional_field': Field(String, is_optional=True)}), is_optional=True
                ),
                inputs=[],
                outputs=[],
                transform_fn=lambda *_args: None,
            )
        ],
    )

    env_type = pipeline_def.environment_type
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['context'].is_optional
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['expectations'].is_optional


def test_required_context_with_required_subfield():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[],
        context_definitions={
            'some_context': PipelineContextDefinition(
                context_fn=lambda *args: None,
                config_field=Field(Dict({'required_field': Field(String)})),
            )
        },
    )

    env_type = pipeline_def.environment_type
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['context'].is_optional is False
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['expectations'].is_optional

    context_union_config_type = env_type.fields['context'].config_type
    assert context_union_config_type.fields['some_context'].is_optional is False


def test_all_optional_field_on_single_context_dict():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[],
        context_definitions={
            'some_context': PipelineContextDefinition(
                context_fn=lambda *args: None,
                config_field=Field(Dict({'optional_field': Field(String, is_optional=True)})),
            )
        },
    )

    env_type = pipeline_def.environment_type
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['context'].is_optional
    assert env_type.fields['execution'].is_optional
    assert env_type.fields['expectations'].is_optional


def test_optional_and_required_context():
    pipeline_def = PipelineDefinition(
        name='some_pipeline',
        solids=[],
        context_definitions={
            'optional_field_context': PipelineContextDefinition(
                context_fn=lambda *args: None,
                config_field=Field(
                    dagster_type=Dict(fields={'optional_field': Field(String, is_optional=True)})
                ),
            ),
            'required_field_context': PipelineContextDefinition(
                context_fn=lambda *args: None,
                config_field=Field(dagster_type=Dict(fields={'required_field': Field(String)})),
            ),
        },
    )

    env_type = pipeline_def.environment_type
    assert env_type.fields['solids'].is_optional
    assert env_type.fields['context'].is_optional is False
    context_type = env_type.fields['context'].config_type

    assert context_type.fields['optional_field_context'].is_optional
    assert context_type.fields['required_field_context'].is_optional

    assert env_type.fields['execution'].is_optional
    assert env_type.fields['expectations'].is_optional

    env_obj = construct_environment_config(
        throwing_evaluate_config_value(
            pipeline_def.environment_type,
            {'context': {'optional_field_context': {'config': {'optional_field': 'foobar'}}}},
        )
    )

    assert env_obj.context.name == 'optional_field_context'
    assert env_obj.context.config == {'optional_field': 'foobar'}


def test_required_inputs():
    @lambda_solid(inputs=[InputDefinition('num', types.Int)], output=OutputDefinition(types.Int))
    def add_one(num):
        return num + 1

    pipeline_def = PipelineDefinition(
        name='required_int_input',
        solids=[add_one],
        dependencies={
            SolidInstance('add_one', 'first_add'): {},
            SolidInstance('add_one', 'second_add'): {'num': DependencyDefinition('first_add')},
        },
    )

    env_type = pipeline_def.environment_type

    solids_type = env_type.fields['solids'].config_type

    first_add_fields = solids_type.fields['first_add'].config_type.fields

    assert 'inputs' in first_add_fields

    inputs_field = first_add_fields['inputs']

    assert inputs_field.is_required

    assert inputs_field.config_type.fields['num'].is_required

    # second_add has a dependency so the input is not available
    assert 'inputs' not in solids_type.fields['second_add'].config_type.fields


def test_mix_required_inputs():
    @lambda_solid(
        inputs=[InputDefinition('left', types.Int), InputDefinition('right', types.Int)],
        output=OutputDefinition(types.Int),
    )
    def add_numbers(left, right):
        return left + right

    @lambda_solid
    def return_three():
        return 3

    pipeline_def = PipelineDefinition(
        name='mixed_required_inputs',
        solids=[add_numbers, return_three],
        dependencies={'add_numbers': {'right': DependencyDefinition('return_three')}},
    )

    solids_type = pipeline_def.environment_type.fields['solids'].config_type
    add_numbers_type = solids_type.fields['add_numbers'].config_type
    inputs_fields_dict = add_numbers_type.fields['inputs'].config_type.fields

    assert 'left' in inputs_fields_dict
    assert not 'right' in inputs_fields_dict



from dagster import check

from dagster.core.execution_context import SystemPipelineExecutionContext
from dagster.core.definitions import InputDefinition, Solid

from dagster.core.errors import DagsterInvariantViolationError

from .objects import (
    ExecutionStep,
    StepOutput,
    StepOutputValue,
    StepKind,
    SingleOutputStepCreationData,
)

INPUT_THUNK_OUTPUT = 'input_thunk_output'


def _create_input_thunk_execution_step(pipeline_context, solid, input_def, input_spec):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    check.invariant(input_def.runtime_type.input_schema)

    def _fn(_step_context, _inputs):
        value = input_def.runtime_type.input_schema.construct_from_config_value(input_spec)
        yield StepOutputValue(output_name=INPUT_THUNK_OUTPUT, value=value)

    return ExecutionStep(
        pipeline_context=pipeline_context,
        key=solid.name + '.' + input_def.name + '.input_thunk',
        step_inputs=[],
        step_outputs=[StepOutput(name=INPUT_THUNK_OUTPUT, runtime_type=input_def.runtime_type)],
        compute_fn=_fn,
        kind=StepKind.INPUT_THUNK,
        solid=solid,
        tags={'input': input_def.name},
    )


def create_input_thunk_execution_step(pipeline_context, solid, input_def, input_spec):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    dependency_structure = pipeline_context.pipeline_def.dependency_structure
    input_handle = solid.input_handle(input_def.name)

    if dependency_structure.has_dep(input_handle):
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'you have specified an input via config while also specifying '
                'a dependency. Either remove the dependency, specify a subdag '
                'to execute, or remove the inputs specification in the environment.'
            ).format(
                pipeline_name=pipeline_context.pipeline.name,
                solid_name=solid.name,
                input_name=input_def.name,
            )
        )

    input_thunk = _create_input_thunk_execution_step(pipeline_context, solid, input_def, input_spec)
    return SingleOutputStepCreationData(input_thunk, INPUT_THUNK_OUTPUT)



from dagster import check
from dagster.core.types.runtime import RuntimeType, resolve_to_runtime_type

from .expectation import ExpectationDefinition
from .utils import check_valid_name, DEFAULT_OUTPUT


class OutputDefinition(object):
    '''An OutputDefinition represents an output from a solid. Solids can have multiple
    outputs. In those cases the outputs must be named. Frequently solids have only one
    output, and so the user can construct a single OutputDefinition that will have
    the default name of "result".

    Attributes:
        runtime_type (DagsterType): Type of the output. Defaults to types.Any.
        name (str): Name of the output. Defaults to "result".
        expectations List[ExpectationDefinition]: Expectations for this output.
        description (str): Description of the output. Optional.
    '''

    def __init__(self, dagster_type=None, name=None, expectations=None, description=None):
        self.name = check_valid_name(check.opt_str_param(name, 'name', DEFAULT_OUTPUT))

        self.runtime_type = check.inst(resolve_to_runtime_type(dagster_type), RuntimeType)

        self.expectations = check.opt_list_param(
            expectations, 'expectations', of_type=ExpectationDefinition
        )
        self.description = check.opt_str_param(description, 'description')

    @property
    def descriptive_key(self):
        return 'output'



'''
This module defines the different notions of context passed to user code during pipeline
execution. There are currently four types of contexts defined:

1) StepExecutionContext
2) TransformExecutionContext
3) ExpectationExecutionContext

They inherit from each other in reverse order. They are also relatively lightweight
objects, and a new one is created for every invocation into user space.

Additionally in order to maintain backwards compatibility (where info.context used
to rule the day) we have a separate class LegacyContextAdapter which can be
easily deleted once we break that API guarantee in 0.4.0.
'''

from abc import ABCMeta, abstractproperty, abstractmethod
from collections import namedtuple
import logging
import warnings
import six
from dagster import check
from dagster.utils.logging import INFO, define_colored_console_logger
from .execution_context import (
    SystemPipelineExecutionContext,
    SystemExpectationExecutionContext,
    SystemStepExecutionContext,
    SystemTransformExecutionContext,
)


class ExecutionContext(namedtuple('_ExecutionContext', 'loggers resources tags')):
    '''
    The user-facing object in the context creation function. The user constructs
    this in order to effect the context creation process. This could be named
    SystemPipelineExecutionContextCreationData although that seemed excessively verbose.
    '''

    def __new__(cls, loggers=None, resources=None, tags=None):
        return super(ExecutionContext, cls).__new__(
            cls,
            loggers=check.opt_list_param(loggers, 'loggers', logging.Logger),
            resources=resources,
            tags=check.opt_dict_param(tags, 'tags'),
        )

    @staticmethod
    def console_logging(log_level=INFO, resources=None):
        return ExecutionContext(
            loggers=[define_colored_console_logger('dagster', log_level)], resources=resources
        )


def _warn_about_context_property():
    warnings.warn(
        (
            'As of 3.0.2 the context property is deprecated. Before your code likely looked '
            'like info.context.some_method_or_property. All of the properties of the nested '
            'context object have been hoisted to the top level, so just rename your info '
            'object to context and access context.some_method_or_property. This property '
            'will be removed in 0.4.0.'
        )
    )


def _warn_about_config_property():
    warnings.warn(
        (
            'As of 3.0.2 the config property is deprecated. Use solid_config instead. '
            'This will be removed in 0.4.0.'
        )
    )


# Delete this after 0.4.0
class LegacyContextAdapter(object):
    __slots__ = ['_pipeline_context']
    '''
    This class mimics the object that was at info.context prior to 0.3.1. It exists in
    order to provide backwards compatibility.
    '''

    def __init__(self, pipeline_context):
        self._pipeline_context = check.inst_param(
            pipeline_context, 'pipeline_context', SystemPipelineExecutionContext
        )

    def has_tag(self, key):
        check.str_param(key, 'key')
        return key in self._pipeline_context.tags

    def get_tag(self, key):
        check.str_param(key, 'key')
        return self._pipeline_context.tags[key]

    @property
    def run_id(self):
        return self._pipeline_context.run_id

    @property
    def event_callback(self):
        return self._pipeline_context.event_callback

    @property
    def has_event_callback(self):
        return self._pipeline_context.event_callback is not None

    @property
    def environment_config(self):
        return self._pipeline_context.environment_config

    @property
    def resources(self):
        return self._pipeline_context.resources

    @property
    def tags(self):
        return self._pipeline_context.tags

    @property
    def persistence_strategy(self):
        return self._pipeline_context.persistence_strategy

    @property
    def events(self):
        return self._pipeline_context.events

    @property
    def log(self):
        return self._pipeline_context.log

    @property
    def pipeline_def(self):
        return self._pipeline_context.pipeline_def

    def debug(self, msg, **kwargs):
        return self._pipeline_context.log.debug(msg, **kwargs)

    def info(self, msg, **kwargs):
        return self._pipeline_context.log.info(msg, **kwargs)

    def warning(self, msg, **kwargs):
        return self._pipeline_context.log.warning(msg, **kwargs)

    def error(self, msg, **kwargs):
        return self._pipeline_context.log.error(msg, **kwargs)

    def critical(self, msg, **kwargs):
        return self._pipeline_context.log.critical(msg, **kwargs)


class StepExecutionContext(object):
    __slots__ = ['_system_step_execution_context', '_legacy_context']

    def __init__(self, system_step_execution_context):
        self._system_step_execution_context = check.inst_param(
            system_step_execution_context,
            'system_step_execution_context',
            SystemStepExecutionContext,
        )
        self._legacy_context = LegacyContextAdapter(system_step_execution_context)

    @property
    def resources(self):
        return self._system_step_execution_context.resources

    @property
    def run_id(self):
        return self._system_step_execution_context.run_id

    @property
    def environment_dict(self):
        return self._system_step_execution_context.environment_dict

    @property
    def pipeline_def(self):
        return self._system_step_execution_context.pipeline_def

    @property
    def log(self):
        return self._system_step_execution_context.log

    @property
    def context(self):
        _warn_about_context_property()
        return self._legacy_context
        # return self._system_step_execution_context.context

    @property
    def solid_def(self):
        return self._system_step_execution_context.solid_def

    @property
    def solid(self):
        return self._system_step_execution_context.solid

    def has_tag(self, key):
        return self._system_step_execution_context.has_tag(key)

    def get_tag(self, key):
        return self._system_step_execution_context.get_tag(key)

    def get_system_context(self):
        '''
        This allows advanced users (e.g. framework authors) to punch through
        to the underlying system context.
        '''
        return self._system_step_execution_context


class AbstractTransformExecutionContext(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    def has_tag(self, key):
        pass

    @abstractmethod
    def get_tag(self, key):
        pass

    @abstractproperty
    def run_id(self):
        pass

    @abstractproperty
    def context(self):
        pass

    @abstractproperty
    def solid_def(self):
        pass

    @abstractproperty
    def solid(self):
        pass

    @abstractproperty
    def pipeline_def(self):
        pass

    @abstractproperty
    def resources(self):
        pass

    @abstractproperty
    def log(self):
        pass


class TransformExecutionContext(StepExecutionContext, AbstractTransformExecutionContext):
    __slots__ = ['_system_transform_execution_context']

    def __init__(self, system_transform_execution_context):
        self._system_transform_execution_context = check.inst_param(
            system_transform_execution_context,
            'system_transform_execution_context',
            SystemTransformExecutionContext,
        )
        super(TransformExecutionContext, self).__init__(system_transform_execution_context)

    @property
    def solid_config(self):
        return self._system_transform_execution_context.solid_config

    @property
    def config(self):
        _warn_about_config_property()
        return self.solid_config


class ExpectationExecutionContext(StepExecutionContext):
    __slots__ = ['_system_expectation_execution_context']

    def __init__(self, system_expectation_execution_context):
        self._system_expectation_execution_context = check.inst_param(
            system_expectation_execution_context,
            'system_expectation_execution_context',
            SystemExpectationExecutionContext,
        )
        super(ExpectationExecutionContext, self).__init__(system_expectation_execution_context)

    @property
    def expectation_def(self):
        return self._system_expectation_execution_context.expectation_def

    @property
    def inout_def(self):
        return self._system_expectation_execution_context.inout_def



from dagster import List, Nullable, Int, PipelineDefinition

from dagster.core.types.runtime import resolve_to_runtime_type, ALL_RUNTIME_BUILTINS


def inner_type_key_set(runtime_type):
    return {t.key for t in runtime_type.inner_types}


def test_inner_types():
    assert resolve_to_runtime_type(Int).inner_types == []

    list_int_runtime = resolve_to_runtime_type(List(Int))
    assert inner_type_key_set(list_int_runtime) == set(['Int'])

    list_list_int_runtime = resolve_to_runtime_type(List(List(Int)))
    assert inner_type_key_set(list_list_int_runtime) == set(['Int', 'List.Int'])

    list_nullable_int_runtime = resolve_to_runtime_type(List(Nullable(Int)))
    assert inner_type_key_set(list_nullable_int_runtime) == set(['Int', 'Nullable.Int'])


def test_display_name():

    int_runtime = resolve_to_runtime_type(Int)
    assert int_runtime.display_name == 'Int'
    list_int_runtime = resolve_to_runtime_type(List(Int))
    assert list_int_runtime.display_name == '[Int]'
    list_list_int_runtime = resolve_to_runtime_type(List(List(Int)))
    assert list_list_int_runtime.display_name == '[[Int]]'
    list_nullable_int_runtime = resolve_to_runtime_type(List(Nullable(Int)))
    assert list_nullable_int_runtime.display_name == '[Int?]'


def test_builtins_available():
    pipeline = PipelineDefinition(name='test_builting_available', solids=[])
    for builtin_type in ALL_RUNTIME_BUILTINS:
        assert pipeline.has_runtime_type(builtin_type.name)
        assert pipeline.runtime_type_named(builtin_type.name).is_builtin



from collections import defaultdict

from dagster import (
    DependencyDefinition,
    Dict,
    Field,
    InputDefinition,
    Int,
    PipelineDefinition,
    RepositoryDefinition,
    String,
    lambda_solid,
    solid,
)


@solid(
    inputs=[InputDefinition('word', String)],
    config_field=Field(Dict({'factor': Field(Int)})),
)
def multiply_the_word(context, word):
    return word * context.solid_config['factor']


@lambda_solid(inputs=[InputDefinition('word')])
def count_letters(word):
    counts = defaultdict(int)
    for letter in word:
        counts[letter] += 1
    return dict(counts)


def define_demo_execution_pipeline():
    return PipelineDefinition(
        name='demo_pipeline',
        solids=[multiply_the_word, count_letters],
        dependencies={
            'count_letters': {
                'word': DependencyDefinition('multiply_the_word')
            }
        },
    )


def define_demo_execution_repo():
    return RepositoryDefinition(
        name='demo_execution_repo',
        pipeline_dict={'demo_pipeline': define_demo_execution_pipeline},
    )



from dagster import check
from .builtin_enum import BuiltinEnum
from .config import Any, ConfigType, List, Nullable
from .field_utils import FieldImpl, FIELD_NO_DEFAULT_PROVIDED, INFER_OPTIONAL_COMPOSITE_FIELD
from .wrapping import WrappingListType, WrappingNullableType


def resolve_to_config_list(list_type):
    check.inst_param(list_type, 'list_type', WrappingListType)
    return List(resolve_to_config_type(list_type.inner_type))


def resolve_to_config_nullable(nullable_type):
    check.inst_param(nullable_type, 'nullable_type', WrappingNullableType)
    return Nullable(resolve_to_config_type(nullable_type.inner_type))


def resolve_to_config_type(dagster_type):
    if dagster_type is None:
        return Any.inst()
    if isinstance(dagster_type, BuiltinEnum):
        return ConfigType.from_builtin_enum(dagster_type)
    ## TODO need to check for runtime type. e.g. List(DataFrame)
    if isinstance(dagster_type, WrappingListType):
        return resolve_to_config_list(dagster_type).inst()
    if isinstance(dagster_type, WrappingNullableType):
        return resolve_to_config_nullable(dagster_type).inst()
    if issubclass(dagster_type, ConfigType):
        return dagster_type.inst()

    check.failed('should not reach')


def Field(
    dagster_type,
    default_value=FIELD_NO_DEFAULT_PROVIDED,
    is_optional=INFER_OPTIONAL_COMPOSITE_FIELD,
    description=None,
):
    return FieldImpl(resolve_to_config_type(dagster_type), default_value, is_optional, description)



from dagster import (
    Dict,
    Field,
    Int,
    PipelineContextDefinition,
    PipelineDefinition,
    SolidDefinition,
    String,
    check,
)

from dagster.core.types import config
from dagster.cli.config_scaffolder import scaffold_pipeline_config, scaffold_type


def fail_me():
    return check.failed('Should not call')


def test_scalars():
    assert scaffold_type(config.Int.inst()) == 0
    assert scaffold_type(config.String.inst()) == ''
    assert scaffold_type(config.Path.inst()) == 'path/to/something'
    assert scaffold_type(config.Bool.inst()) is True
    assert scaffold_type(config.Any.inst()) == 'AnyType'


def test_basic_solids_config():
    pipeline_def = PipelineDefinition(
        name='BasicSolidsConfigPipeline',
        solids=[
            SolidDefinition(
                name='required_field_solid',
                inputs=[],
                outputs=[],
                config_field=Field(Dict(fields={'required_int': Field(Int)})),
                transform_fn=lambda *_args: fail_me(),
            )
        ],
    )

    env_config_type = pipeline_def.environment_type

    assert env_config_type.fields['solids'].is_optional is False
    solids_config_type = env_config_type.fields['solids'].config_type
    assert solids_config_type.fields['required_field_solid'].is_optional is False
    required_solid_config_type = solids_config_type.fields['required_field_solid'].config_type
    assert required_solid_config_type.fields['config'].is_optional is False

    context_config_type = env_config_type.fields['context'].config_type

    assert 'default' in context_config_type.fields
    assert context_config_type.fields['default'].is_optional

    default_context_config_type = context_config_type.fields['default'].config_type

    assert set(default_context_config_type.fields.keys()) == set(
        ['config', 'resources', 'persistence']
    )

    default_context_user_config_type = default_context_config_type.fields['config'].config_type

    assert set(default_context_user_config_type.fields.keys()) == set(['log_level'])

    assert scaffold_pipeline_config(pipeline_def, skip_optional=False) == {
        'context': {
            'default': {'config': {'log_level': ''}, 'persistence': {'file': {}}, 'resources': {}}
        },
        'solids': {'required_field_solid': {'config': {'required_int': 0}}},
        'expectations': {'evaluate': True},
        'execution': {},
    }


def test_two_contexts():
    pipeline_def = PipelineDefinition(
        name='TwoContextsPipeline',
        solids=[],
        context_definitions={
            'context_one': PipelineContextDefinition(
                context_fn=lambda *args: fail_me(),
                config_field=Field(Dict({'context_one_field': Field(String)})),
            ),
            'context_two': PipelineContextDefinition(
                context_fn=lambda *args: fail_me(),
                config_field=Field(Dict({'context_two_field': Field(Int)})),
            ),
        },
    )

    assert scaffold_pipeline_config(pipeline_def) == {'context': {}}

    assert scaffold_pipeline_config(pipeline_def, skip_optional=False) == {
        'context': {
            'context_one': {
                'config': {'context_one_field': ''},
                'persistence': {'file': {}},
                'resources': {},
            },
            'context_two': {
                'config': {'context_two_field': 0},
                'persistence': {'file': {}},
                'resources': {},
            },
        },
        'solids': {},
        'expectations': {'evaluate': True},
        'execution': {},
    }



from abc import ABCMeta, abstractmethod
from contextlib import contextmanager
import os

import six

from dagster import check
from dagster.utils import mkdir_p


class FileStore(six.with_metaclass(ABCMeta)):  # pylint: disable=no-init
    @abstractmethod
    @contextmanager
    def writeable_binary_stream(self, *path_comps):
        pass

    @abstractmethod
    @contextmanager
    def readable_binary_stream(self, *path_comps):
        pass


def check_path_comps(path_comps):
    path_list = check.list_param(list(path_comps), 'path_comps', of_type=str)
    check.param_invariant(path_list, 'path_list', 'Must have at least one comp')
    return path_list


class LocalTempFileStore(FileStore):
    def __init__(self, run_id):
        check.str_param(run_id, 'run_id')
        self.root = os.path.join('/tmp', 'dagster', 'runs', run_id, 'files')
        self._created = False

    def ensure_root_exists(self):
        if not self._created:
            mkdir_p(self.root)

        self._created = True

    @contextmanager
    def writeable_binary_stream(self, *path_comps):
        self.ensure_root_exists()

        path_list = check_path_comps(path_comps)

        target_dir = os.path.join(self.root, *path_list[:-1])
        mkdir_p(target_dir)

        target_path = os.path.join(target_dir, path_list[-1])
        check.invariant(not os.path.exists(target_path))
        with open(target_path, 'wb') as ff:
            yield ff

    @contextmanager
    def readable_binary_stream(self, *path_comps):
        self.ensure_root_exists()

        path_list = check_path_comps(path_comps)

        target_path = os.path.join(self.root, *path_list)
        with open(target_path, 'rb') as ff:
            yield ff

    def has_file(self, *path_comps):
        self.ensure_root_exists()

        path_list = check_path_comps(path_comps)

        target_path = os.path.join(self.root, *path_list)

        if os.path.exists(target_path):
            check.invariant(os.path.isfile(target_path))
            return True
        else:
            return False



import uuid

import pytest

from dagster import (
    Dict,
    ExecutionContext,
    Field,
    OutputDefinition,
    PipelineConfigEvaluationError,
    PipelineContextDefinition,
    PipelineDefinition,
    String,
    execute_pipeline,
    lambda_solid,
    solid,
)


from dagster.utils.logging import INFO

# protected variable. need to test loggers
# pylint: disable=W0212


def test_default_context():
    called = {}

    @solid(inputs=[], outputs=[OutputDefinition()])
    def default_context_transform(context):
        called['yes'] = True
        for logger in context.log.loggers:
            assert logger.level == INFO

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(pipeline)

    assert called['yes']


def test_run_id():
    called = {}

    def construct_context(context):
        called['yes'] = True
        assert uuid.UUID(context.run_id)
        return ExecutionContext()

    pipeline = PipelineDefinition(
        solids=[],
        context_definitions={'default': PipelineContextDefinition(context_fn=construct_context)},
    )
    execute_pipeline(pipeline)

    assert called['yes']


def test_default_context_with_log_level():
    @solid(inputs=[], outputs=[OutputDefinition()])
    def default_context_transform(context):
        for logger in context.log.loggers:
            assert logger.level == INFO

    pipeline = PipelineDefinition(solids=[default_context_transform])
    execute_pipeline(
        pipeline, environment_dict={'context': {'default': {'config': {'log_level': 'INFO'}}}}
    )

    with pytest.raises(PipelineConfigEvaluationError):
        execute_pipeline(
            pipeline, environment_dict={'context': {'default': {'config': {'log_level': 2}}}}
        )


def test_default_value():
    def _get_config_test_solid(config_key, config_value):
        @solid(inputs=[], outputs=[OutputDefinition()])
        def config_test(context):
            assert context.resources == {config_key: config_value}

        return config_test

    pipeline = PipelineDefinition(
        solids=[_get_config_test_solid('field_one', 'heyo')],
        context_definitions={
            'custom_one': PipelineContextDefinition(
                config_field=Field(
                    Dict(
                        {
                            'field_one': Field(
                                dagster_type=String, is_optional=True, default_value='heyo'
                            )
                        }
                    )
                ),
                context_fn=lambda init_context: ExecutionContext(
                    resources=init_context.context_config
                ),
            )
        },
    )

    execute_pipeline(pipeline, environment_dict={'context': {'custom_one': {}}})

    execute_pipeline(pipeline, environment_dict={'context': {'custom_one': None}})


def test_custom_contexts():
    @solid(inputs=[], outputs=[OutputDefinition()])
    def custom_context_transform(context):
        assert context.resources == {'field_one': 'value_two'}

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one': PipelineContextDefinition(
                config_field=Field(Dict({'field_one': Field(dagster_type=String)})),
                context_fn=lambda init_context: ExecutionContext(
                    resources=init_context.context_config
                ),
            ),
            'custom_two': PipelineContextDefinition(
                config_field=Field(Dict({'field_one': Field(dagster_type=String)})),
                context_fn=lambda init_context: ExecutionContext(
                    resources=init_context.context_config
                ),
            ),
        },
    )
    environment_one = {'context': {'custom_one': {'config': {'field_one': 'value_two'}}}}

    execute_pipeline(pipeline, environment_dict=environment_one)

    environment_two = {'context': {'custom_two': {'config': {'field_one': 'value_two'}}}}

    execute_pipeline(pipeline, environment_dict=environment_two)


def test_yield_context():
    events = []

    @solid(inputs=[], outputs=[OutputDefinition()])
    def custom_context_transform(context):
        assert context.resources == {'field_one': 'value_two'}
        assert context.get_tag('foo') == 'bar'  # pylint: disable=W0212
        events.append('during')

    def _yield_context(init_context):
        events.append('before')
        tags = {'foo': 'bar'}
        context = ExecutionContext(resources=init_context.context_config, tags=tags)
        yield context
        events.append('after')

    pipeline = PipelineDefinition(
        solids=[custom_context_transform],
        context_definitions={
            'custom_one': PipelineContextDefinition(
                config_field=Field(Dict({'field_one': Field(dagster_type=String)})),
                context_fn=_yield_context,
            )
        },
    )

    environment_one = {'context': {'custom_one': {'config': {'field_one': 'value_two'}}}}

    execute_pipeline(pipeline, environment_dict=environment_one)

    assert events == ['before', 'during', 'after']


def test_invalid_context():
    @lambda_solid
    def never_transform():
        raise Exception('should never execute')

    default_context_pipeline = PipelineDefinition(
        name='default_context_pipeline', solids=[never_transform]
    )

    environment_context_not_found = {'context': {'not_found': {}}}

    with pytest.raises(
        PipelineConfigEvaluationError, match='Undefined field "not_found" at path root:context'
    ):
        execute_pipeline(default_context_pipeline, environment_dict=environment_context_not_found)

    environment_field_name_mismatch = {'context': {'default': {'config': {'unexpected': 'value'}}}}

    with pytest.raises(PipelineConfigEvaluationError, match='Undefined field "unexpected"'):
        execute_pipeline(default_context_pipeline, environment_dict=environment_field_name_mismatch)

    with_argful_context_pipeline = PipelineDefinition(
        solids=[never_transform],
        context_definitions={
            'default': PipelineContextDefinition(
                config_field=Field(Dict({'string_field': Field(String)})),
                context_fn=lambda init_context: ExecutionContext(resources=init_context.config),
            )
        },
    )

    environment_no_config_error = {'context': {'default': {'config': {}}}}

    with pytest.raises(
        PipelineConfigEvaluationError,
        match=(
            'Error 1: Missing required field  "string_field" at path '
            'root:context:default:config Expected: "{ string_field: String }"'
        ),
    ):
        execute_pipeline(with_argful_context_pipeline, environment_dict=environment_no_config_error)

    environment_type_mismatch_error = {'context': {'default': {'config': {'string_field': 1}}}}

    with pytest.raises(
        PipelineConfigEvaluationError,
        match=(
            'Error 1: Type failure at path "root:context:default:config:string_field" '
            'on type "String". Value at path '
            'root:context:default:config:string_field is not valid. Expected "String".'
        ),
    ):
        execute_pipeline(
            with_argful_context_pipeline, environment_dict=environment_type_mismatch_error
        )



import warnings
import uuid

from dagster import PipelineDefinition, solid, execute_pipeline
from dagster.utils.test import create_test_pipeline_execution_context

# pylint: disable=W0212


def test_noarg_ctor():
    legacy_context = create_test_pipeline_execution_context()
    assert uuid.UUID(legacy_context.run_id)


def test_warning_on_context():
    @solid
    def log_something(context):
        context.context.info('hi')

    pipeline = PipelineDefinition(name='warning_on_context', solids=[log_something])

    with warnings.catch_warnings(record=True) as list_of_warnings:
        result = execute_pipeline(pipeline)
        assert result.success
        assert len(list_of_warnings) == 1
        warning = list_of_warnings[0]
        assert 'As of 3.0.2 the context property is deprecated.' in str(warning.message)


def test_warning_on_config():
    @solid
    def check_config_is_none(context):
        assert context.config is None

    pipeline = PipelineDefinition(name='warning_on_config', solids=[check_config_is_none])

    with warnings.catch_warnings(record=True) as list_of_warnings:
        result = execute_pipeline(pipeline)
        assert result.success
        assert len(list_of_warnings) == 1
        warning = list_of_warnings[0]
        assert 'As of 3.0.2 the config property is deprecated.' in str(warning.message)



from dagster import check
from dagster.utils import single_item

from .builtin_enum import BuiltinEnum
from .config import ConfigType, List, Nullable
from .wrapping import WrappingListType, WrappingNullableType


class InputSchema:
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def construct_from_config_value(self, config_value):
        return config_value


def resolve_config_cls_arg(config_cls):
    if isinstance(config_cls, BuiltinEnum):
        return ConfigType.from_builtin_enum(config_cls)
    elif isinstance(config_cls, WrappingListType):
        return List(resolve_config_cls_arg(config_cls.inner_type))
    elif isinstance(config_cls, WrappingNullableType):
        return Nullable(resolve_config_cls_arg(config_cls.inner_type))
    else:
        check.type_param(config_cls, 'config_cls')
        check.param_invariant(issubclass(config_cls, ConfigType), 'config_cls')
        return config_cls.inst()


def make_bare_input_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)

    class _InputSchema(InputSchema):
        @property
        def schema_type(self):
            return config_type

    return _InputSchema()


class OutputSchema:
    @property
    def schema_type(self):
        check.not_implemented(
            'Must override schema_type in {klass}'.format(klass=type(self).__name__)
        )

    def materialize_runtime_value(self, _config_value, _runtime_value):
        check.not_implemented('Must implement')


def _create_input_schema(config_type, func):
    class _InputSchema(InputSchema):
        @property
        def schema_type(self):
            return config_type

        def construct_from_config_value(self, config_value):
            return func(config_value)

    return _InputSchema()


def input_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)
    return lambda func: _create_input_schema(config_type, func)


def input_selector_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)
    check.param_invariant(config_type.is_selector, 'config_cls')

    def _wrap(func):
        def _selector(config_value):
            selector_key, selector_value = single_item(config_value)
            return func(selector_key, selector_value)

        return _create_input_schema(config_type, _selector)

    return _wrap


def _create_output_schema(config_type, func):
    class _OutputSchema(OutputSchema):
        @property
        def schema_type(self):
            return config_type

        def materialize_runtime_value(self, config_value, runtime_value):
            return func(config_value, runtime_value)

    return _OutputSchema()


def output_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)
    return lambda func: _create_input_schema(config_type, func)


def output_selector_schema(config_cls):
    config_type = resolve_config_cls_arg(config_cls)
    check.param_invariant(config_type.is_selector, 'config_cls')

    def _wrap(func):
        def _selector(config_value, runtime_value):
            selector_key, selector_value = single_item(config_value)
            return func(selector_key, selector_value, runtime_value)

        return _create_output_schema(config_type, _selector)

    return _wrap



import pytest

import dagster.check as check

from dagster import (
    DagsterInvalidDefinitionError,
    DependencyDefinition,
    PipelineDefinition,
    SolidDefinition,
    OutputDefinition,
    InputDefinition,
    solid,
    Field,
    String,
    NamedDict,
)

from dagster.core.utility_solids import define_stub_solid


def solid_a_b_list():
    return [
        SolidDefinition(
            name='A',
            inputs=[],
            outputs=[OutputDefinition()],
            transform_fn=lambda _context, _inputs: None,
        ),
        SolidDefinition(
            name='B',
            inputs=[InputDefinition('b_input')],
            outputs=[],
            transform_fn=lambda _context, _inputs: None,
        ),
    ]


def test_create_pipeline_with_bad_solids_list():
    stub_solid = define_stub_solid('stub', [{'a key': 'a value'}])
    with pytest.raises(check.ParameterCheckError, match='Param "solids" is not a list.'):
        PipelineDefinition(solids=stub_solid, dependencies={})


def test_circular_dep():
    with pytest.raises(DagsterInvalidDefinitionError, match='Circular reference'):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('B')}},
        )


def test_from_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='Solid NOTTHERE in dependency dictionary not found'
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={
                'A': {},
                'B': {'b_input': DependencyDefinition('A')},
                'NOTTHERE': {'b_input': DependencyDefinition('A')},
            },
        )


def test_from_non_existant_input():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='Solid "B" does not have input "not_an_input"'
    ):
        PipelineDefinition(
            solids=solid_a_b_list(), dependencies={'B': {'not_an_input': DependencyDefinition('A')}}
        )


def test_to_solid_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError,
        match='Solid NOTTHERE in DependencyDefinition not found in solid list',
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'A': {}, 'B': {'b_input': DependencyDefinition('NOTTHERE')}},
        )


def test_to_solid_output_not_there():
    with pytest.raises(
        DagsterInvalidDefinitionError, match='Solid A does not have output NOTTHERE'
    ):
        PipelineDefinition(
            solids=solid_a_b_list(),
            dependencies={'B': {'b_input': DependencyDefinition('A', output='NOTTHERE')}},
        )


def test_invalid_item_in_solid_list():
    with pytest.raises(
        DagsterInvalidDefinitionError, match="Invalid item in solid list: 'not_a_solid'"
    ):
        PipelineDefinition(solids=['not_a_solid'])


def test_double_type():
    @solid(config_field=Field(NamedDict('Name', {'some_field': Field(String)})))
    def solid_one(_context):
        raise Exception('should not execute')

    @solid(config_field=Field(NamedDict('Name', {'some_field': Field(String)})))
    def solid_two(_context):
        raise Exception('should not execute')

    with pytest.raises(DagsterInvalidDefinitionError, match='Type names must be unique.'):
        PipelineDefinition(solids=[solid_one, solid_two])



import json

import pytest

from dagster import (
    InputDefinition,
    OutputDefinition,
    PipelineConfigEvaluationError,
    PipelineDefinition,
    Result,
    execute_pipeline,
    lambda_solid,
    solid,
    types,
)

from dagster.utils.test import get_temp_file_name, get_temp_file_names

from dagster.core.definitions.environment_configs import (
    solid_has_config_entry,
    solid_has_configurable_outputs,
)

from dagster.core.execution import create_environment_config, create_execution_plan

from dagster.core.execution_plan.objects import StepOutputHandle

from dagster.core.execution_plan.materialization_thunk import MATERIALIZATION_THUNK_OUTPUT


def single_int_output_pipeline():
    @lambda_solid(output=OutputDefinition(types.Int))
    def return_one():
        return 1

    return PipelineDefinition(name='single_int_output_pipeline', solids=[return_one])


def single_string_output_pipeline():
    @lambda_solid(output=OutputDefinition(types.String))
    def return_foo():
        return 'foo'

    return PipelineDefinition(name='single_string_output_pipeline', solids=[return_foo])


def multiple_output_pipeline():
    @solid(
        outputs=[OutputDefinition(types.Int, 'number'), OutputDefinition(types.String, 'string')]
    )
    def return_one_and_foo(_info):
        yield Result(1, 'number')
        yield Result('foo', 'string')

    return PipelineDefinition(name='multiple_output_pipeline', solids=[return_one_and_foo])


def single_int_named_output_pipeline():
    @lambda_solid(output=OutputDefinition(types.Int, name='named'))
    def return_named_one():
        return Result(1, 'named')

    return PipelineDefinition(name='single_int_named_output_pipeline', solids=[return_named_one])


def no_input_no_output_pipeline():
    @solid(outputs=[])
    def take_nothing_return_nothing(_info):
        pass

    return PipelineDefinition(
        name='no_input_no_output_pipeline', solids=[take_nothing_return_nothing]
    )


def one_input_no_output_pipeline():
    @solid(inputs=[InputDefinition('dummy')], outputs=[])
    def take_input_return_nothing(_info, **_kwargs):
        pass

    return PipelineDefinition(
        name='one_input_no_output_pipeline', solids=[take_input_return_nothing]
    )


def test_solid_has_config_entry():
    pipeline = single_int_output_pipeline()
    assert solid_has_configurable_outputs(pipeline.solid_named('return_one').definition)
    assert solid_has_config_entry(pipeline.solid_named('return_one').definition)


def test_basic_json_default_output_config_schema():
    env = create_environment_config(
        single_int_output_pipeline(),
        {'solids': {'return_one': {'outputs': [{'result': {'json': {'path': 'foo'}}}]}}},
    )

    assert env.solids['return_one']
    assert env.solids['return_one'].outputs == [{'result': {'json': {'path': 'foo'}}}]


def test_basic_json_named_output_config_schema():
    env = create_environment_config(
        single_int_named_output_pipeline(),
        {'solids': {'return_named_one': {'outputs': [{'named': {'json': {'path': 'foo'}}}]}}},
    )

    assert env.solids['return_named_one']
    assert env.solids['return_named_one'].outputs == [{'named': {'json': {'path': 'foo'}}}]


def test_basic_json_misnamed_output_config_schema():
    with pytest.raises(PipelineConfigEvaluationError) as exc_info:
        create_environment_config(
            single_int_named_output_pipeline(),
            {
                'solids': {
                    'return_named_one': {'outputs': [{'wrong_name': {'json': {'path': 'foo'}}}]}
                }
            },
        )

    assert len(exc_info.value.errors) == 1
    assert 'Error 1: Undefined field "wrong_name"' in exc_info.value.message
    assert 'at path root:solids:return_named_one:outputs[0]' in exc_info.value.message


def test_no_outputs_no_inputs_config_schema():
    assert create_environment_config(no_input_no_output_pipeline())

    with pytest.raises(PipelineConfigEvaluationError) as exc_info:
        create_environment_config(no_input_no_output_pipeline(), {'solids': {'return_one': {}}})

    assert len(exc_info.value.errors) == 1
    assert 'Error 1: Undefined field "return_one" at path root:solids' in exc_info.value.message


def test_no_outputs_one_input_config_schema():
    assert create_environment_config(
        one_input_no_output_pipeline(),
        {'solids': {'take_input_return_nothing': {'inputs': {'dummy': {'value': 'value'}}}}},
    )

    with pytest.raises(PipelineConfigEvaluationError) as exc_info:
        create_environment_config(
            one_input_no_output_pipeline(),
            {
                'solids': {
                    'take_input_return_nothing': {
                        'inputs': {'dummy': {'value': 'value'}},
                        'outputs': {},
                    }
                }
            },
        )

    assert len(exc_info.value.errors) == 1
    exp_msg = 'Error 1: Undefined field "outputs" at path root:solids:take_input_return_nothing'
    assert exp_msg in exc_info.value.message


def test_basic_int_execution_plan():
    execution_plan = create_execution_plan(
        single_int_output_pipeline(),
        {'solids': {'return_one': {'outputs': [{'result': {'json': {'path': 'dummy.json'}}}]}}},
    )

    assert len(execution_plan.steps) == 3

    steps = execution_plan.topological_steps()

    assert steps[0].key == 'return_one.transform'
    assert steps[1].key == 'return_one.materialization.output.result.0'
    assert steps[2].key == 'return_one.materialization.output.result.join'


def test_basic_int_json_materialization():
    pipeline = single_int_output_pipeline()

    with get_temp_file_name() as filename:
        result = execute_pipeline(
            pipeline,
            {'solids': {'return_one': {'outputs': [{'result': {'json': {'path': filename}}}]}}},
        )

        assert result.success

        with open(filename, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 1}


def test_basic_string_json_materialization():
    pipeline = single_string_output_pipeline()

    with get_temp_file_name() as filename:
        result = execute_pipeline(
            pipeline,
            {'solids': {'return_foo': {'outputs': [{'result': {'json': {'path': filename}}}]}}},
        )

        assert result.success

        with open(filename, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 'foo'}


def test_basic_int_and_string_execution_plan():
    pipeline = multiple_output_pipeline()
    execution_plan = create_execution_plan(
        pipeline,
        {
            'solids': {
                'return_one_and_foo': {
                    'outputs': [
                        {'string': {'json': {'path': 'dummy_string.json'}}},
                        {'number': {'json': {'path': 'dummy_number.json'}}},
                    ]
                }
            }
        },
    )

    assert len(execution_plan.steps) == 5
    steps = execution_plan.topological_steps()
    assert steps[0].key == 'return_one_and_foo.transform'

    assert_plan_topological_level(
        steps,
        [1, 2],
        [
            'return_one_and_foo.materialization.output.string.0',
            'return_one_and_foo.materialization.output.number.0',
        ],
    )

    assert_plan_topological_level(
        steps,
        [3, 4],
        [
            'return_one_and_foo.materialization.output.number.join',
            'return_one_and_foo.materialization.output.string.join',
        ],
    )

    transform_step = execution_plan.get_step_by_key('return_one_and_foo.transform')

    string_mat_step = execution_plan.get_step_by_key(
        'return_one_and_foo.materialization.output.string.0'
    )
    assert len(string_mat_step.step_inputs) == 1
    assert string_mat_step.step_inputs[0].prev_output_handle == StepOutputHandle.from_step(
        step=transform_step, output_name='string'
    )

    string_mat_join_step = execution_plan.get_step_by_key(
        'return_one_and_foo.materialization.output.string.join'
    )
    assert len(string_mat_join_step.step_inputs) == 1
    assert string_mat_join_step.step_inputs[0].prev_output_handle == StepOutputHandle.from_step(
        step=string_mat_step, output_name=MATERIALIZATION_THUNK_OUTPUT
    )


def test_basic_int_and_string_json_materialization():

    pipeline = multiple_output_pipeline()

    with get_temp_file_names(2) as file_tuple:
        filename_one, filename_two = file_tuple  # pylint: disable=E0632
        result = execute_pipeline(
            pipeline,
            {
                'solids': {
                    'return_one_and_foo': {
                        'outputs': [
                            {'string': {'json': {'path': filename_one}}},
                            {'number': {'json': {'path': filename_two}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, 'r') as ff_1:
            value = json.loads(ff_1.read())
            assert value == {'value': 'foo'}

        with open(filename_two, 'r') as ff_2:
            value = json.loads(ff_2.read())
            assert value == {'value': 1}


def read_file_contents(path):
    with open(path, 'r') as ff:
        return ff.read()


def test_basic_int_and_string_json_multiple_materialization_execution_plan():
    pipeline = multiple_output_pipeline()
    execution_plan = create_execution_plan(
        pipeline,
        {
            'solids': {
                'return_one_and_foo': {
                    'outputs': [
                        {'string': {'json': {'path': 'foo'}}},
                        {'string': {'json': {'path': 'bar'}}},
                        {'number': {'json': {'path': 'baaz'}}},
                        {'number': {'json': {'path': 'quux'}}},
                    ]
                }
            }
        },
    )

    steps = execution_plan.topological_steps()

    assert len(steps) == 7

    assert_plan_topological_level(
        steps,
        [1, 2, 3, 4],
        [
            'return_one_and_foo.materialization.output.number.0',
            'return_one_and_foo.materialization.output.number.1',
            'return_one_and_foo.materialization.output.string.0',
            'return_one_and_foo.materialization.output.string.1',
        ],
    )

    assert_plan_topological_level(
        steps,
        [5, 6],
        [
            'return_one_and_foo.materialization.output.number.join',
            'return_one_and_foo.materialization.output.string.join',
        ],
    )


def test_basic_int_and_string_json_multiple_materialization():

    pipeline = multiple_output_pipeline()

    with get_temp_file_names(4) as file_tuple:
        # False positive for unbalanced tuple unpacking
        # pylint: disable=E0632
        filename_one, filename_two, filename_three, filename_four = file_tuple
        result = execute_pipeline(
            pipeline,
            {
                'solids': {
                    'return_one_and_foo': {
                        'outputs': [
                            {'string': {'json': {'path': filename_one}}},
                            {'string': {'json': {'path': filename_two}}},
                            {'number': {'json': {'path': filename_three}}},
                            {'number': {'json': {'path': filename_four}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 'foo'}

        with open(filename_two, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 'foo'}

        with open(filename_three, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 1}

        with open(filename_four, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 1}


def assert_step_before(steps, first_step, second_step):
    step_keys = [step.key for step in steps]
    assert step_keys.index(first_step) < step_keys.index(second_step)


def assert_plan_topological_level(steps, step_nums, step_keys):
    assert set(steps[step_num].key for step_num in step_nums) == set(step_keys)


def test_basic_int_multiple_serializations_execution_plan():
    execution_plan = create_execution_plan(
        single_int_output_pipeline(),
        {
            'solids': {
                'return_one': {
                    'outputs': [
                        {'result': {'json': {'path': 'dummy_one.json'}}},
                        {'result': {'json': {'path': 'dummy_two.json'}}},
                    ]
                }
            }
        },
    )

    assert len(execution_plan.steps) == 4

    steps = execution_plan.topological_steps()
    assert steps[0].key == 'return_one.transform'

    assert_plan_topological_level(
        steps,
        [1, 2],
        [
            'return_one.materialization.output.result.0',
            'return_one.materialization.output.result.1',
        ],
    )

    assert steps[3].key == 'return_one.materialization.output.result.join'


def test_basic_int_json_multiple_materializations():
    pipeline = single_int_output_pipeline()

    with get_temp_file_names(2) as file_tuple:
        filename_one, filename_two = file_tuple  # pylint: disable=E0632
        result = execute_pipeline(
            pipeline,
            {
                'solids': {
                    'return_one': {
                        'outputs': [
                            {'result': {'json': {'path': filename_one}}},
                            {'result': {'json': {'path': filename_two}}},
                        ]
                    }
                }
            },
        )

        assert result.success

        with open(filename_one, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 1}

        with open(filename_two, 'r') as ff:
            value = json.loads(ff.read())
            assert value == {'value': 1}



from dagster import check

from dagster.core.definitions import (
    solids_in_topological_order,
    InputDefinition,
    OutputDefinition,
    Solid,
    SolidOutputHandle,
)

from dagster.core.errors import DagsterInvariantViolationError

from dagster.core.execution_context import SystemPipelineExecutionContext

from .expectations import create_expectations_subplan, decorate_with_expectations

from .input_thunk import create_input_thunk_execution_step

from .materialization_thunk import decorate_with_output_materializations

from .objects import (
    ExecutionPlan,
    ExecutionStep,
    ExecutionValueSubplan,
    StepInput,
    StepOutputHandle,
    StepKind,
)

from .transform import create_transform_step


class StepOutputMap(dict):
    def __getitem__(self, key):
        check.inst_param(key, 'key', SolidOutputHandle)
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        check.inst_param(key, 'key', SolidOutputHandle)
        check.inst_param(val, 'val', StepOutputHandle)
        return dict.__setitem__(self, key, val)


# This is the state that is built up during the execution plan build process.
# steps is just a list of the steps that have been created
# step_output_map maps logical solid outputs (solid_name, output_name) to particular
# step outputs. This covers the case where a solid maps to multiple steps
# and one wants to be able to attach to the logical output of a solid during execution
class PlanBuilder:
    def __init__(self):
        self.steps = []
        self.step_output_map = StepOutputMap()


def create_execution_plan_core(pipeline_context):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)

    plan_builder = PlanBuilder()

    for solid in solids_in_topological_order(pipeline_context.pipeline_def):

        step_inputs = create_step_inputs(pipeline_context, plan_builder, solid)

        solid_transform_step = create_transform_step(pipeline_context, solid, step_inputs)

        plan_builder.steps.append(solid_transform_step)

        for output_def in solid.definition.output_defs:
            subplan = create_subplan_for_output(
                pipeline_context, solid, solid_transform_step, output_def
            )
            plan_builder.steps.extend(subplan.steps)

            output_handle = solid.output_handle(output_def.name)
            plan_builder.step_output_map[output_handle] = subplan.terminal_step_output_handle

    return create_execution_plan_from_steps(pipeline_context, plan_builder.steps)


def create_execution_plan_from_steps(pipeline_context, steps):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.list_param(steps, 'steps', of_type=ExecutionStep)

    step_dict = {step.key: step for step in steps}
    deps = {step.key: set() for step in steps}

    seen_keys = set()

    for step in steps:
        if step.key in seen_keys:
            keys = [s.key for s in steps]
            check.failed(
                'Duplicated key {key}. Full list: {key_list}.'.format(key=step.key, key_list=keys)
            )

        seen_keys.add(step.key)

        for step_input in step.step_inputs:
            deps[step.key].add(step_input.prev_output_handle.step_key)

    return ExecutionPlan(pipeline_context.pipeline_def, step_dict, deps)


def create_subplan_for_input(pipeline_context, solid, prev_step_output_handle, input_def):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(prev_step_output_handle, 'prev_step_output_handle', StepOutputHandle)
    check.inst_param(input_def, 'input_def', InputDefinition)

    if pipeline_context.environment_config.expectations.evaluate and input_def.expectations:
        return create_expectations_subplan(
            pipeline_context,
            solid,
            input_def,
            prev_step_output_handle,
            kind=StepKind.INPUT_EXPECTATION,
        )
    else:
        return ExecutionValueSubplan.empty(prev_step_output_handle)


def create_subplan_for_output(pipeline_context, solid, solid_transform_step, output_def):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(solid_transform_step, 'solid_transform_step', ExecutionStep)
    check.inst_param(output_def, 'output_def', OutputDefinition)

    subplan = decorate_with_expectations(pipeline_context, solid, solid_transform_step, output_def)

    return decorate_with_output_materializations(pipeline_context, solid, output_def, subplan)


def get_input_source_step_handle(pipeline_context, plan_builder, solid, input_def):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(plan_builder, 'plan_builder', PlanBuilder)
    check.inst_param(solid, 'solid', Solid)
    check.inst_param(input_def, 'input_def', InputDefinition)

    input_handle = solid.input_handle(input_def.name)
    solid_config = pipeline_context.environment_config.solids.get(solid.name)
    dependency_structure = pipeline_context.pipeline_def.dependency_structure
    if solid_config and input_def.name in solid_config.inputs:
        step_creation_data = create_input_thunk_execution_step(
            pipeline_context, solid, input_def, solid_config.inputs[input_def.name]
        )
        plan_builder.steps.append(step_creation_data.step)
        return step_creation_data.step_output_handle
    elif dependency_structure.has_dep(input_handle):
        solid_output_handle = dependency_structure.get_dep(input_handle)
        return plan_builder.step_output_map[solid_output_handle]
    else:
        raise DagsterInvariantViolationError(
            (
                'In pipeline {pipeline_name} solid {solid_name}, input {input_name} '
                'must get a value either (a) from a dependency or (b) from the '
                'inputs section of its configuration.'
            ).format(
                pipeline_name=pipeline_context.pipeline.name,
                solid_name=solid.name,
                input_name=input_def.name,
            )
        )


def create_step_inputs(pipeline_context, plan_builder, solid):
    check.inst_param(pipeline_context, 'pipeline_context', SystemPipelineExecutionContext)
    check.inst_param(plan_builder, 'plan_builder', PlanBuilder)
    check.inst_param(solid, 'solid', Solid)

    step_inputs = []

    for input_def in solid.definition.input_defs:
        prev_step_output_handle = get_input_source_step_handle(
            pipeline_context, plan_builder, solid, input_def
        )

        subplan = create_subplan_for_input(
            pipeline_context, solid, prev_step_output_handle, input_def
        )

        plan_builder.steps.extend(subplan.steps)
        step_inputs.append(
            StepInput(input_def.name, input_def.runtime_type, subplan.terminal_step_output_handle)
        )

    return step_inputs



from dagster import check
from dagster.core.errors import DagsterInvalidDefinitionError, DagsterInvariantViolationError
from .pipeline import PipelineDefinition


class RepositoryDefinition(object):
    '''Define a repository that contains a collection of pipelines.

    Attributes:
        name (str): The name of the pipeline.
        pipeline_dict (Dict[str, callable]):
            An dictionary of pipelines. The value of the dictionary is a function that takes
            no parameters and returns a PipelineDefiniton.

            We pass callables instead of the PipelineDefinitions itself so that they can be
            created on demand when accessed by name.

            As the pipelines are retrieved it ensures that the keys of the dictionary and the
            name of the pipeline are the same.

    '''

    def __init__(self, name, pipeline_dict, enforce_uniqueness=True):
        '''
        Args:
            name (str): Name of pipeline.
            pipeline_dict (Dict[str, callable]): See top-level class documentation
        '''
        self.name = check.str_param(name, 'name')

        check.dict_param(pipeline_dict, 'pipeline_dict', key_type=str)

        for val in pipeline_dict.values():
            check.is_callable(val, 'Value in pipeline_dict must be function')

        self.pipeline_dict = pipeline_dict

        self._pipeline_cache = {}

        self.enforce_uniqueness = enforce_uniqueness

    def has_pipeline(self, name):
        check.str_param(name, 'name')
        return name in self.pipeline_dict

    def get_pipeline(self, name):
        '''Get a pipeline by name. Only constructs that pipeline and caches it.

        Args:
            name (str): Name of the pipeline to retriever

        Returns:
            PipelineDefinition: Instance of PipelineDefinition with that name.
'''
        check.str_param(name, 'name')

        if name in self._pipeline_cache:
            return self._pipeline_cache[name]

        try:
            pipeline = self.pipeline_dict[name]()
        except KeyError:
            raise check.CheckError(
                'Could not find pipeline "{name}". Found: {pipeline_names}.'.format(
                    name=name,
                    pipeline_names=', '.join(
                        [
                            '"{pipeline_name}"'.format(pipeline_name=pipeline_name)
                            for pipeline_name in self.pipeline_dict.keys()
                        ]
                    ),
                )
            )
        check.invariant(
            pipeline.name == name,
            'Name does not match. Name in dict {name}. Name in pipeline {pipeline.name}'.format(
                name=name, pipeline=pipeline
            ),
        )

        self._pipeline_cache[name] = check.inst(
            pipeline,
            PipelineDefinition,
            'Function passed into pipeline_dict with key {key} must return a PipelineDefinition'.format(
                key=name
            ),
        )

        return pipeline

    def iterate_over_pipelines(self):
        '''Yield all pipelines one at a time

        Returns:
            Iterable[PipelineDefinition]:
        '''
        for name in self.pipeline_dict.keys():
            yield self.get_pipeline(name)

    def get_all_pipelines(self):
        '''Return all pipelines as a list

        Returns:
            List[PipelineDefinition]:

        '''
        pipelines = list(self.iterate_over_pipelines())

        self._construct_solid_defs(pipelines)

        return pipelines

    def _construct_solid_defs(self, pipelines):
        solid_defs = {}
        solid_to_pipeline = {}
        for pipeline in pipelines:
            for solid_def in pipeline.solid_defs:
                if solid_def.name not in solid_defs:
                    solid_defs[solid_def.name] = solid_def
                    solid_to_pipeline[solid_def.name] = pipeline.name
                elif self.enforce_uniqueness:
                    if not solid_defs[solid_def.name] is solid_def:
                        raise DagsterInvalidDefinitionError(
                            'Trying to add duplicate solid def {} in {}, Already saw in {}'.format(
                                solid_def.name, pipeline.name, solid_to_pipeline[solid_def.name]
                            )
                        )
        return solid_defs

    def get_solid_def(self, name):
        check.str_param(name, 'name')

        if not self.enforce_uniqueness:
            raise DagsterInvariantViolationError(
                (
                    'In order for get_solid_def to have reliable semantics '
                    'you must construct the repo with ensure_uniqueness=True'
                )
            )

        solid_defs = self._construct_solid_defs(self.get_all_pipelines())

        if name not in solid_defs:
            check.failed('could not find solid_def {}'.format(name))

        return solid_defs[name]

    def solid_def_named(self, name):
        check.str_param(name, 'name')
        for pipeline in self.get_all_pipelines():
            for solid in pipeline.solids:
                if solid.definition.name == name:
                    return solid.definition

        check.failed('Did not find ' + name)



from enum import Enum

from dagster import check


class DagsterExecutionFailureReason(Enum):
    USER_CODE_ERROR = 'USER_CODE_ERROR'
    FRAMEWORK_ERROR = 'FRAMEWORK_ERROR'
    EXPECTATION_FAILURE = 'EXPECATION_FAILURE'


class DagsterError(Exception):
    @property
    def is_user_code_error(self):
        return False


class DagsterUserError(DagsterError):
    pass


class DagsterRuntimeCoercionError(DagsterError):
    '''Runtime checked faild'''


class DagsterInvalidDefinitionError(DagsterUserError):
    '''Indicates that some violation of the definition rules has been violated by the user'''


class DagsterInvariantViolationError(DagsterUserError):
    '''Indicates the user has violated a well-defined invariant that can only be deteremined
    at runtime.
    '''


class DagsterTypeError(DagsterUserError):
    '''Indicates an error in the solid type system (e.g. mismatched arguments)'''


class DagsterUserCodeExecutionError(DagsterUserError):
    '''
    This is base class for any exception that is meant to wrap an Exception
    thrown by user code. It wraps that existing user code. The original_exc_info
    argument to the ctor is meant to be a sys.exc_info at the site of constructor.

    Example:

    output_type = step.step_output_dict[output_name].runtime_type
    try:
        context.persistence_strategy.write_value(
            output_type.serialization_strategy, output['path'], result.step_output_data.value
        )
    except Exception as e:  # pylint: disable=broad-except
        raise_from(
            DagsterExecutionStepExecutionError(...)
            e,
        )

    '''

    def __init__(self, *args, **kwargs):
        # original_exc_info should be gotten from a sys.exc_info() call at the
        # callsite inside of the exception handler. this will allow consuming
        # code to *re-raise* the user error in it's original format
        # for cleaner error reporting that does not have framework code in it
        user_exception = check.inst_param(kwargs.pop('user_exception'), 'user_exception', Exception)
        original_exc_info = check.opt_tuple_param(
            kwargs.pop('original_exc_info', None), 'original_exc_info'
        )
        super(DagsterUserCodeExecutionError, self).__init__(*args, **kwargs)

        self.user_exception = check.opt_inst_param(user_exception, 'user_exception', Exception)
        self.original_exc_info = original_exc_info

    @property
    def is_user_code_error(self):
        return True


class DagsterExecutionStepNotFoundError(DagsterUserError):
    '''
    Generic exception used whenever a user an execution step is specified in an
    api that is consumed by higher level infrastructure (e.g. airflow integration)
    or that needs to be translated into a specific in the GraphQL domain.
    '''

    def __init__(self, *args, **kwargs):
        self.step_key = check.str_param(kwargs.pop('step_key'), 'step_key')
        super(DagsterExecutionStepNotFoundError, self).__init__(*args, **kwargs)


class DagsterExecutionStepExecutionError(DagsterUserCodeExecutionError):
    '''Indicates an error occured during the body of execution step execution'''


class InvalidSubplanErrorData(object):
    def __init__(self, *args, **kwargs):
        from dagster.core.execution_plan.objects import ExecutionStep

        self.pipeline_name = check.str_param(kwargs.pop('pipeline_name'), 'pipeline_name')
        self.step_keys = check.list_param(kwargs.pop('step_keys'), 'step_keys', of_type=str)
        self.step = check.inst_param(kwargs.pop('step'), 'step', ExecutionStep)

        super(InvalidSubplanErrorData, self).__init__(*args, **kwargs)


class DagsterInvalidSubplanMissingInputError(InvalidSubplanErrorData, DagsterUserError):
    '''Indicates that user has attempted to construct an execution subplan
    that cannot be executed because the user needs to specify additional inputs.
    '''

    def __init__(self, *args, **kwargs):
        self.input_name = check.str_param(kwargs.pop('input_name'), 'input_name')
        super(DagsterInvalidSubplanMissingInputError, self).__init__(*args, **kwargs)


class DagsterInvalidSubplanOutputNotFoundError(InvalidSubplanErrorData, DagsterUserError):
    def __init__(self, *args, **kwargs):
        self.output_name = check.str_param(kwargs.pop('output_name'), 'output_name')
        super(DagsterInvalidSubplanOutputNotFoundError, self).__init__(*args, **kwargs)


class DagsterInvalidSubplanInputNotFoundError(InvalidSubplanErrorData, DagsterUserError):
    def __init__(self, *args, **kwargs):
        self.input_name = check.str_param(kwargs.pop('input_name'), 'input_name')
        super(DagsterInvalidSubplanInputNotFoundError, self).__init__(*args, **kwargs)


class DagsterExpectationFailedError(DagsterError):
    '''Thrown with pipeline configured to throw on expectation failure'''

    def __init__(self, expectation_context, value, *args, **kwargs):
        super(DagsterExpectationFailedError, self).__init__(*args, **kwargs)
        self.expectation_context = expectation_context
        self.value = value

    def __repr__(self):
        inout_def = self.expectation_context.inout_def
        return (
            'DagsterExpectationFailedError(solid={solid_name}, {key}={inout_name}, '
            'expectation={e_name}, value={value})'
        ).format(
            solid_name=self.expectation_context.solid.name,
            key=inout_def.descriptive_key,
            inout_name=inout_def.name,
            e_name=self.expectation_context.expectation_def.name,
            value=repr(self.value),
        )

    def __str__(self):
        return self.__repr__()


class DagsterSubprocessExecutionError(DagsterError):
    '''Indicates that an error was encountered when executing part or all of an execution plan
    in a subprocess.
    
    Although this may be a user error, it is intended to be raised in circumstances where we don't
    have access to the sys.exc_info generated by the error (or where the error is raised by a 
    non-Python process).
    '''



class WrappingType(object):
    def __init__(self, inner_type):
        # Cannot check inner_type because of circular references and no fwd declarations
        self.inner_type = inner_type


def List(inner_type):
    return WrappingListType(inner_type)


class WrappingListType(WrappingType):
    pass


def Nullable(inner_type):
    # Cannot check inner_type because of circular references and no fwd declarations
    return WrappingNullableType(inner_type)


class WrappingNullableType(WrappingType):
    pass
